from __future__ import annotations
import weakref
from functools import singledispatch
import numpy as np
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Qt
from typing import Iterator, TYPE_CHECKING, Sequence
from superqt import QToggleSwitch
from magicgui.widgets import Container

from himena.standards import roi
from himena.consts import StandardType
from himena.qt import drag_command
from himena.qt.magicgui import get_type_map
from himena.utils.ndobject import NDObjectCollection
from himena_builtins.qt.widgets._image_components import _roi_items
from himena_builtins.qt.widgets._dragarea import QDraggableArea

if TYPE_CHECKING:
    from typing import Self
    from himena_builtins.qt.widgets.image import QImageView


@singledispatch
def _roi_to_qroi(r: roi.RoiModel) -> _roi_items.QRoi:
    raise ValueError(f"Unsupported ROI type: {type(r)}")


@_roi_to_qroi.register
def _(r: roi.LineRoi) -> _roi_items.QRoi:
    return _roi_items.QLineRoi(r.x1 + 0.5, r.y1 + 0.5, r.x2 + 0.5, r.y2 + 0.5)


@_roi_to_qroi.register
def _(r: roi.RectangleRoi) -> _roi_items.QRoi:
    return _roi_items.QRectangleRoi(r.x, r.y, r.width, r.height)


@_roi_to_qroi.register
def _(r: roi.EllipseRoi) -> _roi_items.QRoi:
    return _roi_items.QEllipseRoi(r.x, r.y, r.width, r.height)


@_roi_to_qroi.register
def _(r: roi.SegmentedLineRoi) -> _roi_items.QRoi:
    return _roi_items.QSegmentedLineRoi(r.xs + 0.5, r.ys + 0.5)


@_roi_to_qroi.register
def _(r: roi.PolygonRoi) -> _roi_items.QRoi:
    roi = _roi_items.QPolygonRoi(r.xs + 0.5, r.ys + 0.5)
    path = roi.path()
    path.closeSubpath()
    roi.setPath(path)
    return roi


@_roi_to_qroi.register
def _(r: roi.RotatedRectangleRoi) -> _roi_items.QRoi:
    return _rotated_roi_to_qroi(r, _roi_items.QRotatedRectangleRoi)


@_roi_to_qroi.register
def _(r: roi.RotatedEllipseRoi) -> _roi_items.QRoi:
    return _rotated_roi_to_qroi(r, _roi_items.QRotatedEllipseRoi)


def _rotated_roi_to_qroi(r: roi.RotatedRectangleRoi | roi.RotatedEllipseRoi, cls):
    xstart, ystart = r.start
    xend, yend = r.end
    return cls(
        QtCore.QPointF(xstart + 0.5, ystart + 0.5),
        QtCore.QPointF(xend + 0.5, yend + 0.5),
        width=r.width,
    )


@_roi_to_qroi.register
def _(r: roi.PointRoi2D) -> _roi_items.QRoi:
    return _roi_items.QPointRoi(r.x + 0.5, r.y + 0.5)


@_roi_to_qroi.register
def _(r: roi.PointsRoi2D) -> _roi_items.QRoi:
    return _roi_items.QPointsRoi(r.xs + 0.5, r.ys + 0.5)


@_roi_to_qroi.register
def _(r: roi.CircleRoi) -> _roi_items.QRoi:
    return _roi_items.QCircleRoi(r.x + 0.5, r.y + 0.5, r.radius)


def from_standard_roi(r: roi.RoiModel, pen: QtGui.QPen) -> _roi_items.QRoi:
    """Convert a standard ROI to a QRoi."""
    out = _roi_to_qroi(r)
    return out.withPen(pen).withLabel(r.name)


Indices = tuple[int, ...]


class QSimpleRoiCollection(QtW.QWidget):
    drag_requested = QtCore.Signal(list)  # list[int] of selected indices
    roi_update_requested = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self._qroi_list = NDObjectCollection[_roi_items.QRoi]()
        self._pen = QtGui.QPen(QtGui.QColor(238, 238, 0), 2)
        self._pen.setCosmetic(True)
        layout = QtW.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self._list_view = QRoiListView(self)

        layout.addWidget(self._list_view, 100, alignment=Qt.AlignmentFlag.AlignTop)
        self._list_view.drag_requested.connect(self.drag_requested.emit)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self)} ROIs)"

    def layout(self) -> QtW.QVBoxLayout:
        return super().layout()

    def extend_from_standard_roi_list(
        self,
        rois: NDObjectCollection[roi.RoiModel],
    ) -> Self:
        """Extend the collection from a list of himena standard ROIs."""
        qrois = np.fromiter(
            (from_standard_roi(r, self._pen) for r in rois),
            dtype=np.object_,
        )
        self.extend(
            NDObjectCollection[_roi_items.QRoi](
                items=qrois,
                indices=rois.indices,
                axis_names=rois.axis_names,
            )
        )
        return self

    def to_standard_roi_list(
        self,
        selections: list[int] | None = None,
    ) -> NDObjectCollection[roi.RoiModel]:
        if selections is None:
            all_rois = self._qroi_list
        else:
            all_rois = self._qroi_list.filter_by_selection(selections)
        return all_rois.map_elements(lambda qroi: qroi.toRoi(), into=roi.RoiListModel)

    def add(self, indices: Indices, roi: _roi_items.QRoi):
        """Add a ROI on the given slice."""
        self._list_view.model().beginInsertRows(
            QtCore.QModelIndex(), len(self._qroi_list), len(self._qroi_list)
        )
        self._qroi_list.add_item(indices, roi)
        self._list_view.model().endInsertRows()

    def extend(self, other: NDObjectCollection[_roi_items.QRoi]):
        self._list_view.model().beginInsertRows(
            QtCore.QModelIndex(),
            len(self._qroi_list),
            len(self._qroi_list) + len(other),
        )
        self._qroi_list.extend(other)
        self._list_view.model().endInsertRows()

    def clear(self):
        self._list_view.model().beginResetModel()
        self._qroi_list.clear()
        self._list_view.model().endResetModel()

    def set_selections(self, selections: list[int]):
        sel_model = self._list_view.selectionModel()
        sel_model.clear()
        model = self._list_view.model()
        for i in selections:
            sel_model.select(
                model.index(i, 0), QtCore.QItemSelectionModel.SelectionFlag.Select
            )

    def selections(self) -> list[int]:
        """List of selected indices"""
        return [idx.row() for idx in self._list_view.selectionModel().selectedIndexes()]

    def __getitem__(self, key: int) -> _roi_items.QRoi:
        return self._qroi_list[key]

    def count(self) -> int:
        return len(self._qroi_list)

    __len__ = count

    def __iter__(self) -> Iterator[_roi_items.QRoi]:
        yield from self._qroi_list

    def get_rois_on_slice(self, indices: tuple[int, ...]) -> Sequence[_roi_items.QRoi]:
        """Return a list of ROIs on the given slice."""
        out = self._qroi_list.filter_by_indices(indices).items
        return out

    def index_in_slice(self, indices: Indices, ith: int) -> int:
        """Return the `index`-th ROI in the slice `indices`."""
        mask = self._qroi_list.mask_by_indices(indices)
        if mask is None:
            index_total = ith
        else:
            index_total = np.where(mask)[0][ith]
        return index_total

    def pop_roi_in_slice(self, indices: Indices, ith: int) -> _roi_items.QRoi:
        """Pop the `index`-th ROI in the slice `indices`."""
        index_total = self.index_in_slice(indices, ith)
        qindex = self._list_view.model().index(index_total)
        self._list_view.model().beginRemoveRows(qindex, index_total, index_total)
        roi = self._qroi_list.pop(index_total)
        self._list_view.model().endRemoveRows()
        self._list_view.update()
        return roi

    def pop_rois(self, indices: list[int]):
        sl = np.ones(self.count(), dtype=bool)
        sl[indices] = False
        self._list_view.model().beginRemoveRows(QtCore.QModelIndex(), 0, len(sl) - 1)
        self._qroi_list = self._qroi_list.filter_by_selection(sl)
        self._list_view.model().endRemoveRows()

    def flatten_roi(self, indices: int | list[int]) -> None:
        return self.flatten_roi_along(indices, slice(None))

    def flatten_roi_along(self, indices: int | list[int], axis) -> None:
        self._qroi_list.indices[indices, axis] = -1
        return None

    def move_roi(self, indices: int | list[int], new_dims: tuple[int, ...]) -> None:
        if not isinstance(indices, list):
            indices = [indices]
        for i in indices:
            self._qroi_list.indices[i, :] = new_dims
        self.roi_update_requested.emit()
        return None

    def remove_selected_rois(self):
        self.pop_rois(self.selections())
        self.roi_update_requested.emit()


class QRoiCollection(QSimpleRoiCollection):
    """Object to store and manage multiple ROIs in nD images."""

    show_rois_changed = QtCore.Signal(bool)
    show_labels_changed = QtCore.Signal(bool)
    roi_item_clicked = QtCore.Signal(tuple, _roi_items.QRoi)
    key_pressed = QtCore.Signal(QtGui.QKeyEvent)
    key_released = QtCore.Signal(QtGui.QKeyEvent)

    def __init__(self, parent: QImageView):
        super().__init__(parent)
        self._image_view_ref = weakref.ref(parent)
        self._list_view.clicked.connect(self._on_item_clicked)
        self._list_view.key_pressed.connect(self.key_pressed)
        self._list_view.key_released.connect(self.key_released)

        self.layout().addWidget(self._list_view, 100, Qt.AlignmentFlag.AlignTop)
        self._dragarea = QDraggableArea()
        self._dragarea.setToolTip("Drag the image ROIs")
        self._dragarea.setFixedSize(14, 14)
        self._add_btn = QtW.QPushButton("+")
        self._add_btn.setToolTip("Register current ROI to the list")
        self._add_btn.setFixedSize(14, 14)
        self._add_btn.setSizePolicy(
            QtW.QSizePolicy.Policy.Minimum, QtW.QSizePolicy.Policy.Minimum
        )
        self._add_btn.clicked.connect(parent._img_view.add_current_roi)
        self._remove_btn = QtW.QPushButton("-")
        self._remove_btn.setToolTip("Remove selected ROI from the list")
        self._remove_btn.setFixedSize(14, 14)
        self._remove_btn.setSizePolicy(
            QtW.QSizePolicy.Policy.Minimum, QtW.QSizePolicy.Policy.Minimum
        )
        _btn_layout = QtW.QHBoxLayout()
        _btn_layout.setContentsMargins(0, 0, 0, 0)
        _btn_layout.setSpacing(1)
        _btn_layout.addWidget(self._dragarea, alignment=Qt.AlignmentFlag.AlignLeft)
        _btn_layout.addWidget(QtW.QWidget(), 100)
        _btn_layout.addWidget(self._add_btn, alignment=Qt.AlignmentFlag.AlignRight)
        _btn_layout.addWidget(self._remove_btn, alignment=Qt.AlignmentFlag.AlignRight)
        self.layout().addLayout(_btn_layout)
        self._dragarea.dragged.connect(self._on_dragged)
        self._roi_visible_btn = QToggleSwitch()
        self._roi_visible_btn.setText("Show ROIs")
        self._roi_visible_btn.setChecked(True)
        self._roi_labels_btn = QToggleSwitch()
        self._roi_labels_btn.setText("Labels")
        self._roi_labels_btn.setChecked(True)
        self._roi_visible_btn.setSizePolicy(
            QtW.QSizePolicy(
                QtW.QSizePolicy.Policy.Minimum, QtW.QSizePolicy.Policy.Minimum
            )
        )
        self._roi_labels_btn.setSizePolicy(
            QtW.QSizePolicy(
                QtW.QSizePolicy.Policy.Minimum, QtW.QSizePolicy.Policy.Minimum
            )
        )
        self.layout().addWidget(
            self._roi_visible_btn, alignment=Qt.AlignmentFlag.AlignBottom
        )
        self.layout().addWidget(
            self._roi_labels_btn, alignment=Qt.AlignmentFlag.AlignBottom
        )

        self.show_rois_changed.connect(parent._img_view.set_show_rois)
        self.show_labels_changed.connect(parent._img_view.set_show_labels)
        self.key_pressed.connect(parent.keyPressEvent)
        self.key_released.connect(parent.keyReleaseEvent)
        self.roi_item_clicked.connect(self._roi_item_clicked)
        self._add_btn.clicked.connect(parent._img_view.add_current_roi)
        self.drag_requested.connect(parent._on_drag_roi_requested)
        self._remove_btn.clicked.connect(self.remove_selected_rois)

        self._roi_visible_btn.toggled.connect(self._on_roi_visible_btn_clicked)
        self._roi_labels_btn.toggled.connect(self._on_roi_labels_btn_clicked)

        self.setToolTip("List of ROIs in the image")

    def _roi_item_clicked(self, indices: tuple[int, ...], qroi: _roi_items.QRoi):
        view = self._image_view_ref()
        if view is None:
            return

        # find the proper slider indices
        if (ninds := len(indices)) < (ndim_rem := view._dims_slider.count()):
            # this happens when the ROI is flattened
            if ninds > 0:
                higher_dims = ndim_rem - ninds
                indices_filled = view._dims_slider.value()[:higher_dims] + indices
                view._dims_slider.set_value_no_emit(indices_filled)
                view._slider_changed(view._dims_slider.value(), force_sync=True)
        else:
            view._dims_slider.set_value_no_emit(indices)
            view._slider_changed(view._dims_slider.value(), force_sync=True)

        # update selection in the image viewer
        view._img_view.select_item(qroi, is_registered_roi=True)

    def _on_roi_visible_btn_clicked(self, checked: bool):
        if self._roi_labels_btn.isChecked() and not checked:
            self._roi_labels_btn.setChecked(False)
        self.show_rois_changed.emit(checked)

    def _on_roi_labels_btn_clicked(self, checked: bool):
        if checked and not self._roi_visible_btn.isChecked():
            self._roi_visible_btn.setChecked(True)
        self.show_labels_changed.emit(checked)

    def _on_item_clicked(self, index: QtCore.QModelIndex):
        r = index.row()
        if 0 <= r < len(self._qroi_list):
            qroi = self._qroi_list[r]
            indices = self._qroi_list.indices[r]
            self.roi_item_clicked.emit(tuple(indices), qroi)

    def _on_dragged(self):
        _s = "" if len(self._qroi_list) == 1 else "s"
        return drag_command(
            self._image_view_ref(),
            command_id="builtins:rois:duplicate",
            type=StandardType.ROIS,
            desc=f"{len(self._qroi_list)} ROI{_s}",
        )


class QRoiListView(QtW.QListView):
    # NOTE: list view usually has a focus. Key events have to be forwarded.
    key_pressed = QtCore.Signal(QtGui.QKeyEvent)
    key_released = QtCore.Signal(QtGui.QKeyEvent)
    drag_requested = QtCore.Signal(list)  # list[int] of selected indices

    def __init__(self, parent: QRoiCollection):
        super().__init__(parent)
        self.setModel(QRoiListModel(parent))
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setSelectionMode(QtW.QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setEditTriggers(QtW.QAbstractItemView.EditTrigger.EditKeyPressed)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._show_context_menu)
        self.setDragEnabled(True)
        self.setMouseTracking(True)
        self._hover_drag_indicator = QDraggableArea(self)
        self._hover_drag_indicator.setWindowFlags(Qt.WindowType.FramelessWindowHint)
        self._hover_drag_indicator.setFixedSize(14, 14)
        self._hover_drag_indicator.hide()
        self._hover_drag_indicator.dragged.connect(self._on_drag)
        self._indicator_index: int = -1

    def set_selection(self, indices: list[int]):
        """Set the selection of the list view."""
        self.clearSelection()
        for i in indices:
            index = self.model().index(i, 0)
            self.selectionModel().select(
                index, QtCore.QItemSelectionModel.SelectionFlag.Select
            )
        if len(indices) > 0:
            self.parent()._on_item_clicked(self.model().index(indices[0], 0))

    def _show_context_menu(self, point):
        index_under_cursor = self.indexAt(point)
        if not index_under_cursor.isValid():
            return
        menu = self._prep_context_menu(index_under_cursor)
        menu.exec(self.mapToGlobal(point))

    def _prep_context_menu(self, index: QtCore.QModelIndex):
        _qroi_list = self.parent()._qroi_list
        menu = QtW.QMenu(self)
        selected_indices = self.parent().selections()
        action_rename = menu.addAction("Rename", lambda: self.edit(index))
        action_rename.setToolTip("Rename the selected ROI")
        menu_flatten = menu.addMenu("Flatten Along ...")

        action_flatten = menu_flatten.addAction(
            "All axes", lambda: self.parent().flatten_roi(selected_indices)
        )
        menu_flatten.addSeparator()
        for i in range(_qroi_list.ndim):
            _action = menu_flatten.addAction(
                f"{i}: {_qroi_list.axis_names[i]}",
                lambda i=i: self.parent().flatten_roi_along(selected_indices, [i]),
            )
            _action.setToolTip(
                f"Flatten the selected ROI along {_qroi_list.axis_names[i]!r} axis"
            )
            if self.parent()._qroi_list.indices[index.row(), i] == -1:
                _action.setEnabled(False)  # already flattened
        action_flatten.setToolTip("Flatten the selected ROI into 2D")
        action_move = menu.addAction(
            "Move To ...",
            lambda: self._move_to(selected_indices),
        )
        action_move.setToolTip("Move the selected ROI to another dimension")
        action_delete = menu.addAction(
            "Delete", lambda: self.parent().remove_selected_rois()
        )
        action_delete.setToolTip("Delete the selected ROIs")
        return menu

    def _move_to(self, indices):
        if dims := self._select_dimensions_from_dialog():
            self.parent().move_roi(indices, dims)

    def _make_dialog(self, options: dict[str, dict]) -> tuple[QtW.QDialog, Container]:
        mgui_typemap = get_type_map()
        dialog = QtW.QDialog(self)
        dialog.setWindowTitle("Select dimensions")
        layout = QtW.QVBoxLayout(dialog)
        container = Container()
        container.margins = (0, 0, 0, 0)
        for name, each in options.items():
            annotation = each.pop("annotation", None)
            widget = mgui_typemap.create_widget(
                annotation=annotation,
                name=name,
                options=each,
            )
            container.append(widget)
        layout.addWidget(container.native)
        button_layout = QtW.QHBoxLayout()
        button_layout.setContentsMargins(0, 0, 0, 0)
        button_layout.setSpacing(2)
        button_layout.setAlignment(Qt.AlignmentFlag.AlignRight)
        ok_button = QtW.QPushButton("OK")
        cancel_button = QtW.QPushButton("Cancel")
        button_layout.addWidget(ok_button)
        button_layout.addWidget(cancel_button)
        layout.addLayout(button_layout)
        ok_button.clicked.connect(dialog.accept)
        cancel_button.clicked.connect(dialog.reject)
        return dialog, container

    def _options_for_select_dimensions(self) -> dict[str, dict]:
        options = {}
        img_view = self.parent()._image_view_ref()
        if img_view is None:
            return None
        slider_dims = img_view._dims_slider.maximums()
        slider_values = list(img_view._dims_slider.value())
        for ith, axis in enumerate(self.parent()._qroi_list.axis_names):
            options[axis] = {
                "value": slider_values[ith],
                "annotation": int,
                "widget_type": "Slider",
                "min": 0,
                "max": slider_dims[ith],
            }
        return options

    def _select_dimensions_from_dialog(self) -> tuple[int, ...] | None:
        options = self._options_for_select_dimensions()
        dialog, container = self._make_dialog(options)
        if dialog.exec() == QtW.QDialog.DialogCode.Accepted:
            return tuple(each.value for each in container)
        else:
            return None

    def keyPressEvent(self, a0: QtGui.QKeyEvent):
        if a0.key() == Qt.Key.Key_F2:
            self.edit(self.currentIndex())
            return
        elif a0.key() in (Qt.Key.Key_Up, Qt.Key.Key_Down):
            return super().keyPressEvent(a0)
        self.key_pressed.emit(a0)

    def keyReleaseEvent(self, a0):
        self.key_released.emit(a0)
        return super().keyReleaseEvent(a0)

    def mouseMoveEvent(self, e):
        if e.button() == Qt.MouseButton.NoButton:
            self._hover_event(e.pos())
        return super().mouseMoveEvent(e)

    def _hover_event(self, pos: QtCore.QPoint):
        index = self.indexAt(pos)
        if index.isValid():
            self.setCursor(Qt.CursorShape.PointingHandCursor)
            index_rect = self.rectForIndex(index)
            top_right = index_rect.topRight()
            top_right.setX(top_right.x() - 14)
            self._hover_drag_indicator.move(top_right)
            self._hover_drag_indicator.show()
            self._indicator_index = index.row()
        else:
            self.setCursor(Qt.CursorShape.ArrowCursor)
            self._hover_drag_indicator.hide()

    def leaveEvent(self, a0):
        self._hover_drag_indicator.hide()
        return super().leaveEvent(a0)

    def sizeHint(self):
        return QtCore.QSize(180, 900)  # set to a very large value to make it expanded

    def parent(self) -> QRoiCollection:
        return super().parent()

    def _on_drag(self) -> None:
        sels = self.parent().selections()
        if self._indicator_index not in sels:
            sels.append(self._indicator_index)
            index = self.model().index(self._indicator_index, 0)
            self.selectionModel().select(
                index, QtCore.QItemSelectionModel.SelectionFlag.Select
            )
        self.drag_requested.emit(sels)


_FLAGS = (
    Qt.ItemFlag.ItemIsEnabled
    | Qt.ItemFlag.ItemIsSelectable
    | Qt.ItemFlag.ItemIsEditable
)


class QRoiListModel(QtCore.QAbstractListModel):
    """The list model used for displaying ROIs."""

    def __init__(self, col: QRoiCollection, parent=None):
        super().__init__(parent)
        self._col = col

    def rowCount(self, parent=None):
        return len(self._col)

    def data(self, index: QtCore.QModelIndex, role: int):
        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            r = index.row()
            if 0 <= r < len(self._col):
                return self._col[r].label()
            return None
        elif role == Qt.ItemDataRole.DecorationRole:
            r = index.row()
            if 0 <= r < len(self._col):
                pixmap = QtGui.QPixmap(24, 24)
                pixmap.fill(Qt.GlobalColor.black)
                return self._col[r].makeThumbnail(pixmap).scaled(
                    12, 12, Qt.AspectRatioMode.KeepAspectRatio,
                    Qt.TransformationMode.SmoothTransformation
                )  # fmt: skip
        elif role == Qt.ItemDataRole.FontRole:
            font = self._col.font()
            font.setPointSize(10)
            if index == self._col._list_view.currentIndex():
                font.setBold(True)
            return font
        elif role == Qt.ItemDataRole.SizeHintRole:
            return QtCore.QSize(80, 14)
        elif role == Qt.ItemDataRole.ToolTipRole:
            r = index.row()
            if 0 <= r < len(self._col):
                _indices = tuple(self._col._qroi_list.indices[r])
                _type = self._col._qroi_list[r]._roi_type()
                if len(_indices) > 0:
                    return f"{_type.title()} ROI on slice {_indices}"
                else:
                    return f"{_type.title()} ROI"
        return None

    def flags(self, index):
        return _FLAGS

    def setData(self, index, value, role):
        if role == Qt.ItemDataRole.EditRole:
            self._col._qroi_list[index.row()].set_label(value)
            return True



from __future__ import annotations

from qtpy import QtWidgets as QtW, QtCore, QtGui


class QRenameLineEdit(QtW.QLineEdit):
    rename_requested = QtCore.Signal(str)

    def __init__(self, parent: QtW.QWidget):
        super().__init__(parent)
        self.setHidden(True)

        @self.editingFinished.connect
        def _():
            if not self.isVisible():
                return
            self.setHidden(True)
            text = self.text()
            if text:
                self.rename_requested.emit(text)

    def keyPressEvent(self, a0: QtGui.QKeyEvent) -> None:
        if a0.key() == QtCore.Qt.Key.Key_Escape:
            self.setHidden(True)
        return super().keyPressEvent(a0)

    def focusOutEvent(self, a0: QtGui.QFocusEvent) -> None:
        self.setHidden(True)
        return super().focusOutEvent(a0)


class QTabRenameLineEdit(QRenameLineEdit):
    """LineEdit for renaming tab in QTabWidget."""

    def __init__(self, parent: QtW.QTabWidget, allow_duplicate: bool = True):
        super().__init__(parent)
        self._current_edit_index: int | None = None
        self._allow_duplicate = allow_duplicate

        @self.rename_requested.connect
        def _(new_name: str):
            if self._current_edit_index is None:
                return
            if not self._allow_duplicate:
                for i in range(self.parent().count()):
                    if i == self._current_edit_index:
                        continue
                    if self.parent().tabText(i) == new_name:
                        raise ValueError(f"Duplicate tab name: {new_name!r}")
            parent.setTabText(self._current_edit_index, new_name)

        parent.currentChanged.connect(self._hide_me)
        parent.tabBarDoubleClicked.connect(self.start_edit)

    def parent(self) -> QtW.QTabWidget:
        return super().parent()

    def keyPressEvent(self, a0: QtGui.QKeyEvent) -> None:
        if a0.key() == QtCore.Qt.Key.Key_Escape:
            self.setHidden(True)
        return super().keyPressEvent(a0)

    def showEvent(self, a0: QtGui.QShowEvent | None) -> None:
        self._current_edit_index = self.parent().currentIndex()
        return super().showEvent(a0)

    def _move_line_edit(
        self,
        rect: QtCore.QRect,
        text: str,
    ) -> QtW.QLineEdit:
        geometry = self.geometry()
        geometry.setWidth(rect.width())
        geometry.setHeight(rect.height())
        geometry.moveCenter(rect.center())
        geometry.adjust(4, 4, -2, -2)
        self.setGeometry(geometry)
        self.setText(text)
        self.setHidden(False)
        self.setFocus()
        self.selectAll()

    def _hide_me(self):
        self.setHidden(True)

    def start_edit(self, index: int):
        """Enter edit table name mode."""
        rect = self._tab_rect(index)
        self._current_edit_index = index
        self._move_line_edit(rect, self.parent().tabText(index))
        return None

    def _tab_rect(self, index: int) -> QtCore.QRect:
        """Get QRect of the tab at index."""
        tab_widget = self.parent()
        rect = tab_widget.tabBar().tabRect(index)
        return rect



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence, TypeVar
from logging import getLogger
from himena.consts import StandardType
from himena.types import WidgetDataModel
from himena.standards import BaseMetadata
from himena.profile import AppProfile, load_app_profile
from himena.workflow import ProgrammaticMethod

if TYPE_CHECKING:
    import numpy as np
    from himena.widgets import MainWindow
    from himena.standards.model_meta import ImageChannel, DimAxis
    from himena.standards.roi import RoiModel, RoiListModel

_LOGGER = getLogger(__name__)
_T = TypeVar("_T")


def new_window(
    profile: str | AppProfile | None = None,
    *,
    plugins: Sequence[str] | None = None,
    backend: str = "qt",
    app_attributes: dict[str, Any] = {},
) -> MainWindow:
    """Create a new window with the specified profile and additional plugins."""
    from himena._app_model import get_model_app
    from himena.widgets._initialize import init_application
    from himena.plugins.install import (
        install_plugins,
        override_keybindings,
        install_default_configs,
    )

    plugins = list(plugins or [])

    # NOTE: the name of AppProfile and the app_model.Application must be the same.
    if isinstance(profile, str):
        app_prof = load_app_profile(profile)
    elif isinstance(profile, AppProfile):
        app_prof = profile
    elif profile is None:
        app_prof = load_app_profile("default", create_default=True)
    else:
        raise TypeError("`profile` must be a str or an AppProfile object.")
    model_app = get_model_app(app_prof.name)
    model_app.attributes.update(dict(app_attributes))
    plugins = [p for p in plugins if p not in app_prof.plugins]  # filter duplicates
    plugins = app_prof.plugins + plugins
    _init_backend(backend)
    install_default_configs()
    results = install_plugins(model_app, plugins)

    # create the main window
    init_application(model_app)
    override_keybindings(model_app, app_prof)
    main_window = _get_main_window(backend, model_app, theme=app_prof.theme)

    # execute startup commands (don't raise exceptions, just log them)
    exceptions: list[tuple[str, dict, Exception]] = []
    for cmd, kwargs in app_prof.startup_commands:
        try:
            main_window.exec_action(cmd, with_params=kwargs)
        except Exception as e:
            exceptions.append((cmd, kwargs, e))
    if exceptions:
        _LOGGER.error("Exceptions occurred during startup commands:")
        for cmd, kwargs, exc in exceptions:
            _LOGGER.error("  %r (parameters=%r): %s", cmd, kwargs, exc)

    # call plugin startup functions
    for result in results:
        result.call_startup(main_window)
    return main_window


def _init_backend(backend: str):
    # This function needed for proper import time calculation
    if backend == "qt":
        import himena.qt  # noqa: F401


def create_model(
    value: _T,
    *,
    type: str | None = None,
    title: str | None = None,
    extension_default: str | None = None,
    extensions: list[str] | None = None,
    metadata: object | None = None,
    force_open_with: str | None = None,
    editable: bool = True,
    add_empty_workflow: bool = False,
) -> WidgetDataModel[_T]:
    """Helper function to create a WidgetDataModel."""
    if type is None:
        if isinstance(metadata, BaseMetadata):
            type = metadata.expected_type()
    if type is None:
        type = StandardType.ANY
    out = WidgetDataModel(
        value=value,
        type=type,
        title=title,
        extension_default=extension_default,
        extensions=extensions or [],
        metadata=metadata,
        force_open_with=force_open_with,
        editable=editable,
    )
    if add_empty_workflow:
        out.workflow = ProgrammaticMethod().construct_workflow()
    return out


def create_text_model(
    value: str,
    *,
    title: str | None = None,
    language: str | None = None,
    spaces: int = 4,
    selection: tuple[int, int] | None = None,
    font_family: str | None = None,
    font_size: int = 10,
    encoding: str | None = None,
    extension_default: str | None = None,
    extensions: list[str] | None = None,
    force_open_with: str | None = None,
    editable: bool = True,
) -> WidgetDataModel[str]:
    """Helper function to create a WidgetDataModel for text."""
    from himena.standards.model_meta import TextMeta

    return create_model(
        value,
        title=title,
        metadata=TextMeta(
            language=language,
            spaces=spaces,
            selection=selection,
            font_family=font_family,
            font_size=font_size,
            encoding=encoding,
        ),
        extension_default=extension_default,
        extensions=extensions or [],
        force_open_with=force_open_with,
        editable=editable,
    )


def create_table_model(
    value: Any,
    *,
    title: str | None = None,
    current_position: tuple[int, int] | None = None,
    selections: list[tuple[tuple[int, int], tuple[int, int]]] | None = None,
    separator: str | None = None,
    extension_default: str | None = None,
    extensions: list[str] | None = None,
    force_open_with: str | None = None,
    editable: bool = True,
) -> WidgetDataModel[np.ndarray[tuple[int, int], str]]:
    """Helper function to create a WidgetDataModel for tables."""
    from himena.standards.model_meta import TableMeta

    return create_model(
        value,
        title=title,
        metadata=TableMeta(
            current_position=current_position,
            selections=selections or [],
            separator=separator,
        ),
        extension_default=extension_default,
        extensions=extensions or [],
        force_open_with=force_open_with,
        editable=editable,
    )


def create_dataframe_model(
    value: _T,
    *,
    title: str | None = None,
    current_position: tuple[int, int] | None = None,
    selections: list[tuple[tuple[int, int], tuple[int, int]]] | None = None,
    separator: str | None = None,
    transpose: bool = False,
    extension_default: str | None = None,
    extensions: list[str] | None = None,
    force_open_with: str | None = None,
    editable: bool = True,
) -> WidgetDataModel[_T]:
    """Helper function to create a WidgetDataModel for dataframes."""
    from himena.standards.model_meta import DataFrameMeta

    return create_model(
        value,
        title=title,
        metadata=DataFrameMeta(
            current_position=current_position,
            selections=selections or [],
            separator=separator,
            transpose=transpose,
        ),
        extension_default=extension_default,
        extensions=extensions or [],
        force_open_with=force_open_with,
        editable=editable,
    )


def create_array_model(
    value: _T,
    *,
    axes: list[str | dict | DimAxis] | None = None,
    current_indices: list[int] | None = None,
    selections: list[tuple[tuple[int, int], tuple[int, int]]] | None = None,
    unit: str | None = None,
    title: str | None = None,
    extension_default: str | None = None,
    extensions: list[str] | None = None,
    force_open_with: str | None = None,
    editable: bool = True,
) -> WidgetDataModel[_T]:
    """Helper function to create a WidgetDataModel for arrays."""
    from himena.standards.model_meta import ArrayMeta

    return create_model(
        value,
        title=title,
        metadata=ArrayMeta(
            axes=_norm_axes(axes, value),
            selections=selections or [],
            unit=unit,
            current_indices=current_indices,
        ),
        extension_default=extension_default,
        extensions=extensions or [],
        force_open_with=force_open_with,
        editable=editable,
    )


def create_image_model(
    value: _T,
    *,
    axes: list[str | dict | DimAxis] | None = None,
    channels: list[str | dict | ImageChannel] | None = None,
    channel_axis: int | None = None,
    current_roi: RoiModel | None = None,
    current_roi_index: int | None = None,
    rois: RoiListModel | None = None,
    contrast_limits: tuple[float, float] | None = None,
    current_indices: list[int | None] | None = None,
    is_rgb: bool = False,
    unit: str | None = None,
    title: str | None = None,
    extension_default: str | None = None,
    extensions: list[str] | None = None,
    force_open_with: str | None = None,
    editable: bool = True,
) -> WidgetDataModel[_T]:
    """Helper function to create a WidgetDataModel for images."""
    from himena.standards.model_meta import ImageMeta
    from himena.standards.roi import RoiListModel

    return create_model(
        value,
        title=title,
        metadata=ImageMeta(
            axes=_norm_axes(axes, value),
            channels=_norm_channels(channels),
            is_rgb=is_rgb,
            current_roi=current_roi,
            current_roi_index=current_roi_index,
            rois=rois or RoiListModel(),
            contrast_limits=contrast_limits,
            current_indices=current_indices,
            channel_axis=channel_axis,
            unit=unit,
        ),
        extension_default=extension_default,
        extensions=extensions or [],
        force_open_with=force_open_with,
        editable=editable,
    )


def _get_main_window(backend: str, *args, **kwargs) -> MainWindow:
    if backend == "qt":
        from himena.qt import MainWindowQt

        return MainWindowQt(*args, **kwargs)
    elif backend == "mock":
        from himena.mock import MainWindowMock

        return MainWindowMock(*args, **kwargs)
    raise ValueError(f"Invalid backend: {backend}")


def _norm_axes(axes, value) -> list[DimAxis]:
    from himena.data_wrappers import wrap_array
    from himena.standards.model_meta import DimAxis

    if axes is None:
        arr = wrap_array(value)
        _axes = arr.infer_axes()
    else:
        _axes: list[DimAxis] = []
        for axis in axes:
            if isinstance(axis, str):
                _axes.append(DimAxis(name=axis))
            elif isinstance(axis, dict):
                _axes.append(DimAxis(**axis))
            elif isinstance(axis, DimAxis):
                _axes.append(axis)
            else:
                raise TypeError(f"Invalid axis type: {type(axis)}")
    return _axes


def _norm_channels(channels) -> list[ImageChannel] | None:
    from himena.standards.model_meta import ImageChannel

    if channels is None:
        _channels = [ImageChannel.default()]
    else:
        _channels = []
        for ch in channels:
            if isinstance(ch, str):
                _channels.append(ImageChannel(colormap=ch))
            elif isinstance(ch, ImageChannel):
                _channels.append(ch)
            else:
                raise TypeError(f"Invalid channel type: {type(ch)}")
    return _channels



from dataclasses import asdict, dataclass, field
import math
from pathlib import Path
import timeit
import uuid
from typing import (
    Any,
    Callable,
    ClassVar,
    Literal,
    NamedTuple,
    NewType,
    TypeAlias,
    TypeVar,
    Generic,
    TYPE_CHECKING,
)
from pydantic import BaseModel, Field, field_validator
from himena._descriptors import SaveBehavior
from himena.workflow import (
    Workflow,
    LocalReaderMethod,
    CommandExecution,
    parse_parameter,
    ModelParameter,
    WindowParameter,
)
from himena.utils.enum import StrEnum
from himena.utils.misc import is_subtype
from himena.consts import PYDANTIC_CONFIG_STRICT

if TYPE_CHECKING:
    from typing import Self
    from himena.utils.misc import PluginInfo


class DockArea(StrEnum):
    """Area of the dock widget."""

    TOP = "top"
    BOTTOM = "bottom"
    LEFT = "left"
    RIGHT = "right"


class WindowState(StrEnum):
    """State of the sub window."""

    MIN = "min"
    MAX = "max"
    NORMAL = "normal"
    FULL = "full"


DockAreaString: TypeAlias = Literal["top", "bottom", "left", "right"]
WindowStateString: TypeAlias = Literal["min", "max", "normal", "full"]


class NewWidgetBehavior(StrEnum):
    """Behavior of adding a widget."""

    TAB = "tab"
    WINDOW = "window"


_T = TypeVar("_T")
_U = TypeVar("_U")

if TYPE_CHECKING:

    class GenericModel(Generic[_T], BaseModel):
        pass
else:

    class GenericModel(BaseModel):
        def __class_getitem__(cls, item):
            return cls


class _Void:
    pass


_void = _Void()


class WidgetDataModel(GenericModel[_T]):
    """A data model that represents a widget containing an internal data.

    Attributes
    ----------
    value : Any
        Internal value.
    type : str, optional
        Type of the internal data. Type hierarchy is separated by dots. For example,
        "text.plain" is a subtype of "text".
    title : str, optional
        Title for the widget. If not given, the title will be generated from the source
        path when this model is added to the GUI.
    extension_default : str, optional
        Default file extension for saving. This is used when the user saves the data
        without specifying the file extension.
    extensions : list[str], optional
        List of allowed file extensions to save this data.
    metadata : Any, optional
        Metadata that may be used for storing additional information of the internal
        data or describing the state of the widget.
    workflow : WorkflowList, optional
        History of how this data is created.
    force_open_with : str, optional
        Force open with a specific plugin if given.
    """

    model_config = PYDANTIC_CONFIG_STRICT

    value: _T = Field(..., description="Internal value.")
    type: str = Field(..., description="Type of the internal data.")
    title: str | None = Field(
        default=None,
        description="Default title for the widget.",
    )
    extension_default: str | None = Field(
        default=None,
        description="Default file extension for saving.",
    )
    extensions: list[str] = Field(
        default_factory=list,
        description="List of allowed file extensions.",
    )
    metadata: object | None = Field(
        default=None,
        description="Metadata that may be used for storing additional information of "
        "the internal data or describing the state of the widget.",
    )  # fmt: skip
    workflow: Workflow = Field(default_factory=Workflow)
    force_open_with: str | None = Field(
        default=None,
        description="Force open with a specific plugin if given.",
    )
    save_behavior_override: SaveBehavior | None = Field(
        default=None,
        description="Override the default save behavior.",
    )
    editable: bool = Field(True, description="Whether the widget is editable.")
    update_inplace: bool = Field(
        False,
        description="Whether to update the input data instead of creating a new window.",
    )
    window_rect_override: Callable[["Size"], "WindowRect"] | None = Field(None)

    def with_value(
        self,
        value: _U,
        type: str | None = None,
        *,
        title: str | None = None,
        metadata: object | None = _void,
        save_behavior_override: SaveBehavior | _Void | None = _void,
        update_inplace: bool = False,
    ) -> "WidgetDataModel[_U]":
        """Return a model with the new value."""
        update = {"value": value}
        if type is not None:
            update["type"] = type
        if metadata is not _void:
            update["metadata"] = metadata
        if title is not None:
            update["title"] = title
        if save_behavior_override is not _void:
            update["save_behavior_override"] = save_behavior_override
        update.update(
            workflow=Workflow(),
            force_open_with=None,
            update_inplace=update_inplace,
        )  # these parameters must be reset
        return self.model_copy(update=update)

    def astype(self, new_type: str):
        update = {"type": new_type}
        return self.model_copy(update=update)

    def _with_source(
        self,
        source: str | Path | list[str | Path],
        plugin: "PluginInfo | None" = None,
        id: uuid.UUID | None = None,
    ) -> "WidgetDataModel[_T]":
        """Return a new instance with the source path."""
        if plugin is None:
            plugin_name = None
        else:
            plugin_name = plugin.to_str()
        if isinstance(source, list):
            path = [Path(s).resolve() for s in source]
        else:
            path = Path(source).resolve()
        wf = LocalReaderMethod(
            path=path,
            plugin=plugin_name,
            output_model_type=self.type,
            id=id or uuid.uuid4(),
        ).construct_workflow()
        to_update = {"workflow": wf}
        if self.title is None:
            if isinstance(path, list):
                to_update.update({"title": "File group"})
            else:
                to_update.update({"title": path.name})
        return self.model_copy(update=to_update)

    def with_open_plugin(
        self,
        open_with: str,
        *,
        workflow: Workflow | _Void | None = _void,
        save_behavior_override: SaveBehavior | _Void | None = _void,
    ) -> "WidgetDataModel[_T]":
        update = {"force_open_with": open_with}
        if workflow is not _void:
            update["workflow"] = workflow
        if save_behavior_override is not _void:
            update["save_behavior_override"] = save_behavior_override
        return self.model_copy(update=update)

    def with_metadata(
        self,
        metadata: Any,
        update_inplace: bool = False,
    ) -> "WidgetDataModel[_T]":
        """Return a new instance with the given metadata."""
        update = {"metadata": metadata, "update_inplace": update_inplace}
        return self.model_copy(update=update)

    def write_to_directory(
        self,
        directory: str | Path,
        *,
        plugin: str | None = None,
    ) -> Path:
        from himena import _providers

        ins = _providers.WriterStore.instance()
        title = self.title or "Untitled"
        path = Path(directory) / title
        if path.suffix == "":
            if ext := self.extension_default:
                path = path.with_suffix(ext)
            elif exts := self.extensions:
                path = path.with_suffix(exts[0])
            else:
                raise ValueError("Could not determine the file extension.")
        ins.run(self, path, min_priority=0, plugin=plugin)
        return path

    @property
    def source(self) -> Path | list[Path] | None:
        """The direct source path of the data."""
        if isinstance(step := self.workflow.last(), LocalReaderMethod):
            return step.path
        return None

    def is_subtype_of(self, supertype: str) -> bool:
        """Check if the type is a subtype of the given type."""
        return is_subtype(self.type, supertype)

    def with_title_numbering(self, copy: bool = False) -> "WidgetDataModel[_T]":
        """Add [n] suffix to the title."""
        title = self.title
        if title is None:
            title = "Untitled"
        if "." in title:
            stem, ext = title.rsplit(".", 1)
            ext = f".{ext}"
        else:
            stem = title
            ext = ""
        if (
            (last_part := stem.rsplit(" ", 1)[-1]).startswith("[")
            and last_part.endswith("]")
            and last_part[1:-1].isdigit()
        ):
            nth = int(last_part[1:-1])
            stem = stem.rsplit(" ", 1)[0] + f" [{nth + 1}]"
        else:
            stem = stem + " [1]"
        new_title = stem + ext
        if copy:
            return self.model_copy(update={"title": new_title})
        else:
            self.title = new_title
            return self

    @field_validator("extension_default", mode="after")
    def _validate_extension_default(cls, v: str, values):
        if v is None:
            return None
        if not v.startswith("."):
            return f".{v}"
        return v

    @field_validator("extensions", mode="before")
    def _validate_extensions(cls, v):
        if isinstance(v, str):
            v = [v]
        if not all(isinstance(ext, str) for ext in v):
            raise TypeError(f"Invalid type for `extensions`: {type(v)}")
        return [s if s.startswith(".") else f".{s}" for s in v]

    def __repr__(self):
        value_repr = f"<{type(self.value).__name__}>"
        if isinstance(source := self.source, Path):
            source_repr = source.as_posix()
        elif isinstance(source, list):
            if len(source) > 0:
                source_repr = f"[{source[0].as_posix()}, ...]"
            else:
                source_repr = "[]"
        else:
            source_repr = None
        return (
            f"{self.__class__.__name__}(value={value_repr}, source={source_repr}, "
            f"type={self.type!r}, title={self.title!r})"
        )


class ClipboardDataModel(BaseModel):
    """Data model for a clipboard data."""

    model_config = PYDANTIC_CONFIG_STRICT

    text: str | None = Field(
        default=None,
        description="Text in the clipboard if exists.",
    )
    html: str | None = Field(
        default=None,
        description="HTML in the clipboard if exists.",
    )
    image: Any | None = Field(
        default=None,
        description="Image in the clipboard if exists.",
    )
    files: list[Path] = Field(
        default_factory=list,
        description="List of file paths in the clipboard if exists.",
    )
    internal_data: Any | None = Field(
        default=None,
        description="Application specific data in the clipboard if exists. This data "
        "cannot be used across application, but is useful to send Python object to "
        "other widgets.",
    )

    def with_internal_data(self, internal_data) -> "ClipboardDataModel":
        return self.model_copy(update={"internal_data": internal_data})


class DragDataModel(BaseModel):
    model_config = PYDANTIC_CONFIG_STRICT
    getter: Callable[[], WidgetDataModel] | WidgetDataModel = Field(
        ..., description="Getter function to get the data model."
    )
    type: str | None = Field(None, description="Type of the internal data.")

    def data_model(self) -> WidgetDataModel:
        if isinstance(self.getter, WidgetDataModel):
            model = self.getter
        else:
            model = self.getter()
        return model

    def widget_accepts_me(self, widget: Any) -> bool:
        """Return true if the widget accepts this data model to be dropped."""
        if hasattr(widget, "allowed_drop_types"):
            types = widget.allowed_drop_types()
            if self.type is None:
                return True  # not specified. Just allow it.
            if self.type in types:
                return True
        elif hasattr(widget, "dropped_callback"):
            return True
        return False


_V = TypeVar("_V", int, float)


@dataclass(frozen=True)
class Size(Generic[_V]):
    """Size use for any place."""

    width: _V
    height: _V

    def __iter__(self):
        """Iterate over the field to make this class tuple-like."""
        return iter((self.width, self.height))

    def __getitem__(self, index: int):
        if index == 0:
            return self.width
        elif index == 1:
            return self.height
        raise IndexError(f"Index {index!r} out of range.")

    def with_width(self, width: _V) -> "Size[_V]":
        return Size(width, self.height)

    def with_height(self, height: _V) -> "Size[_V]":
        return Size(self.width, height)


@dataclass(frozen=True)
class Rect(Generic[_V]):
    """Rectangle use for any place."""

    left: _V
    top: _V
    width: _V
    height: _V

    @property
    def right(self):
        return self.left + self.width

    @property
    def bottom(self):
        return self.top + self.height

    def with_left(self, left: _V) -> "Rect[_V]":
        return Rect(left, self.top, self.width, self.height)

    def with_top(self, top: _V) -> "Rect[_V]":
        return Rect(self.left, top, self.width, self.height)

    def with_width(self, width: _V) -> "Rect[_V]":
        return Rect(self.left, self.top, width, self.height)

    def with_height(self, height: _V) -> "Rect[_V]":
        return Rect(self.left, self.top, self.width, height)

    def with_right(self, right: _V) -> "Rect[_V]":
        return Rect(right - self.width, self.top, self.width, self.height)

    def with_bottom(self, bottom: _V) -> "Rect[_V]":
        return Rect(self.left, bottom - self.height, self.width, self.height)

    def __iter__(self):
        """Iterate over the field to make this class tuple-like."""
        return iter((self.left, self.top, self.width, self.height))

    def size(self) -> Size[_V]:
        return Size(self.width, self.height)

    def adjust_to_int(
        self,
        how: Literal["inner", "outer"] = "inner",
    ) -> "Rect[int]":
        right = self.right
        bottom = self.bottom
        if how == "inner":
            left = int(math.ceil(self.left))
            top = int(math.ceil(self.top))
            right = int(math.floor(right))
            bottom = int(math.floor(bottom))
        else:
            left = int(math.floor(self.left))
            top = int(math.floor(self.top))
            right = int(math.ceil(right))
            bottom = int(math.ceil(bottom))
        return Rect(left, top, right - left, bottom - top)

    def limit_to(self, xmax: _T, ymax: _T) -> "Rect[_T]":
        """Limit the size of the Rect to the given maximum size."""
        left = max(self.left, 0)
        top = max(self.top, 0)
        right = min(self.right, xmax)
        bottom = min(self.bottom, ymax)
        return Rect(left, top, right - left, bottom - top)

    def move_top_left(self, left: _V, top: _V) -> "Rect[_V]":
        return Rect(left, top, self.width, self.height)

    def move_top_right(self, right: _V, top: _V) -> "Rect[_V]":
        return Rect(right - self.width, top, self.width, self.height)

    def move_bottom_left(self, left: _V, bottom: _V) -> "Rect[_V]":
        return Rect(left, bottom - self.height, self.width, self.height)

    def move_bottom_right(self, right: _V, bottom: _V) -> "Rect[_V]":
        return Rect(right - self.width, bottom - self.height, self.width, self.height)


@dataclass(frozen=True)
class WindowRect(Rect[int]):
    """Rectangle of a window."""

    @classmethod
    def from_tuple(cls, left, top, width, height) -> "WindowRect":
        return cls(int(left), int(top), int(width), int(height))

    def align_left(self, area_size: Size[int]) -> "WindowRect":
        return WindowRect(0, self.top, self.width, self.height)

    def align_right(self, area_size: Size[int]) -> "WindowRect":
        w0, _ = area_size
        return WindowRect(w0 - self.width, self.top, self.width, self.height)

    def align_top(self, area_size: Size[int]) -> "WindowRect":
        return WindowRect(self.left, 0, self.width, self.height)

    def align_bottom(self, area_size: Size[int]) -> "WindowRect":
        _, h0 = area_size
        return WindowRect(self.left, h0 - self.height, self.width, self.height)

    def align_center(self, area_size: Size[int]) -> "WindowRect":
        w0, h0 = area_size
        return WindowRect(
            (w0 - self.width) / 2,
            (h0 - self.height) / 2,
            self.width,
            self.height,
        )

    def resize_relative(self, wratio: float, hratio: float) -> "WindowRect":
        if wratio <= 0 or hratio <= 0:
            raise ValueError("Ratios must be positive.")
        return WindowRect(
            self.left,
            self.top,
            round(self.width * wratio),
            round(self.height * hratio),
        )


@dataclass(frozen=True)
class Margins(Generic[_V]):
    left: _V
    top: _V
    right: _V
    bottom: _V

    def __iter__(self):
        """Iterate over the field to make this class tuple-like."""
        return iter((self.left, self.top, self.right, self.bottom))

    @classmethod
    def from_rects(cls, inner: Rect[_V], outer: Rect[_V]) -> "Margins[_V]":
        """Calculate the margins from the inner and outer rectangles."""
        return cls(
            inner.left - outer.left,
            inner.top - outer.top,
            outer.right - inner.right,
            outer.bottom - inner.bottom,
        )


class _HasDynamicAttribute:
    _ATTR_NAME: str

    @classmethod
    def get(cls, obj) -> "Self | None":
        return getattr(obj, cls._ATTR_NAME, None)

    def set(self, obj) -> None:
        setattr(obj, self._ATTR_NAME, self)


@dataclass
class GuiConfiguration(_HasDynamicAttribute):
    """Configuration for parametric widget (interpreted by the injection processor)"""

    _ATTR_NAME: ClassVar[str] = "__himena_gui_config__"

    title: str | None = None
    preview: bool = False
    auto_close: bool = True
    show_parameter_labels: bool = True
    run_async: bool = False
    result_as: Literal["window", "below", "right"] = "window"

    def asdict(self) -> dict[str, Any]:
        """Return the configuration as a dictionary."""
        return asdict(self)


@dataclass(frozen=True)
class ModelTrack(_HasDynamicAttribute):
    """Model to track how model is created."""

    _ATTR_NAME: ClassVar[str] = "__himena_model_track__"

    command_id: str
    contexts: list[ModelParameter | WindowParameter] = field(default_factory=list)
    workflow: Workflow = field(default_factory=Workflow)
    time_start: float = field(default=0.0)

    def to_workflow(self, parameters: dict[str, Any]) -> Workflow:
        """Construct a workflow based on the given parameters."""
        params = []
        more_workflows: list[Workflow] = []
        for k, v in parameters.items():
            if k == "is_previewing":
                continue
            param, wf = parse_parameter(k, v)
            params.append(param)
            more_workflows.append(wf)
        workflow = Workflow.concat([self.workflow] + more_workflows)
        return workflow.with_step(
            CommandExecution(
                command_id=self.command_id,
                contexts=self.contexts,
                parameters=params,
                execution_time=timeit.default_timer() - self.time_start,
            )
        )


@dataclass(frozen=True)
class FutureInfo(_HasDynamicAttribute):
    _ATTR_NAME: ClassVar[str] = "__himena_future_info__"

    type_hint: Any
    track: ModelTrack | None = None
    kwargs: dict[str, Any] = field(default_factory=dict)
    top_left: tuple[int, int] | None = None
    size: Size[int] | None = None

    def resolve_type_hint(self, ns: dict[str, Any]) -> "FutureInfo":
        if isinstance(self.type_hint, str):
            typ = ns.get(self.type_hint)
            if typ is None:
                raise ValueError(f"Could not resolve the type hint: {self.type_hint}")
            type_hint = typ
        else:
            type_hint = self.type_hint
        return FutureInfo(
            type_hint=type_hint,
            track=self.track,
            kwargs=self.kwargs,
            top_left=self.top_left,
            size=self.size,
        )


Parametric = NewType("Parametric", Any)
"""Callback for a parametric function.

This type can be interpreted by the injection store processor. For example, in the
following code, `my_plugin_function` will be converted into a parametric widget
with inputs `a` and `b`..

``` python
from himena.plugin import register_function
@register_function(...)
def my_plugin_function(...) -> Parametric:
    def callback_func(a: int, b: str) -> WidgetDataModel:
        ...
    return my_plugin_function
```
"""


class ParametricWidgetProtocol:
    """Protocol used for return annotation of a parametric widget."""

    def __new__(cls, *args, **kwargs) -> None:
        if cls is ParametricWidgetProtocol:
            raise TypeError("ParametricWidgetProtocol cannot be instantiated.")
        return super().__new__(cls)

    def get_output(self, *args, **kwargs) -> Any:
        raise NotImplementedError


class BackendInstructions(BaseModel):
    """Instructions for the backend that are only relevant to user interface."""

    model_config = PYDANTIC_CONFIG_STRICT

    animate: bool = Field(
        default=True,
        description="Whether to animate",
        frozen=True,
    )
    confirm: bool = Field(
        default=True,
        description="Whether to show a confirmation dialog",
        frozen=True,
    )
    choose_one_dialog_response: Callable[[], Any] | None = Field(
        default=None,
        description="If provided, choose-one dialog will be skipped and this function "
        "will be called to get the response.",
        frozen=True,
    )
    file_dialog_response: Callable[[], Any] | None = Field(
        default=None,
        description="If provided, file dialog will be skipped and this function will "
        "be called to get the response.",
        frozen=True,
    )
    gui_execution: bool = Field(default=True)
    process_model_output: bool = Field(default=True)
    unwrap_future: bool = Field(default=False)

    def updated(self, **kwargs) -> "BackendInstructions":
        return self.model_copy(update=kwargs)


class WidgetClassTuple(NamedTuple):
    """Class for storing registered widget class."""

    type: str
    widget_class: "type | Callable"  # factory function
    priority: int = 100
    widget_id: str | None = None


WidgetType = NewType("WidgetType", object)
"""This type is used for the return annotation.

``` python
from himena.plugin import register_function
@register_function(...)
def my_plugin_function() -> WidgetType:
    return MyWidget()
```
"""

WidgetConstructor = NewType("WidgetConstructor", object)
"""This type is used for the return annotation.

``` python
from himena.plugin import register_function
@register_function(...)
def my_plugin_function() -> WidgetConstructor:
    return MyWidget
```
"""


@dataclass(frozen=True)
class DropResult:
    """Model that can be returned by `dropped_callback` protocol.

    Attributes
    ----------
    delete_input : bool
        Whether to delete the input data if drop succeeded.
    command_id : str | None
        Command that will be executed when the drop succeeded.
    with_params : dict[str, object] | None
        Parameters that will be passed to the command.
    """

    delete_input: bool = False
    command_id: str | None = None
    with_params: dict[str, object] | None = None



from app_model.types import KeyBindingRule, KeyCode, KeyMod
from himena.consts import MenuId
from himena.widgets import MainWindow
from himena._app_model.actions._registry import ACTIONS


CMD_GROUP = "command@00"


@ACTIONS.append_from_fn(
    id="show-command-palette",
    title="Command palette",
    menus=[{"id": MenuId.TOOLS, "group": CMD_GROUP}],
    keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.KeyP)],
    recording=False,
)
def show_command_palette(ui: MainWindow) -> None:
    """Open the command palette."""
    ui._backend_main_window._show_command_palette("general")


@ACTIONS.append_from_fn(
    id="repeat-command",
    title="Repeat last command",
    menus=[{"id": MenuId.TOOLS, "group": CMD_GROUP}],
    keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyCode.KeyR)],
    recording=False,
)
def repeat_last_command(ui: MainWindow) -> None:
    """Repeat the command that was executed last."""
    if id := ui._history_command.get_from_last(1):
        if action := ui.model_app.registered_actions.get(id):
            ctx = ui._ctx_keys.dict()
            if action.enablement is None or action.enablement.eval(ctx):
                ui.exec_action(id)



from __future__ import annotations
from typing import Literal

from qtpy import QtWidgets as QtW, QtCore, QtGui
from himena.consts import MonospaceFontFamily


class QSplitterHandle(QtW.QSplitterHandle):
    def __init__(self, parent: QtW.QSplitter, side: Literal["left", "right"] = "right"):
        super().__init__(QtCore.Qt.Orientation.Horizontal, parent)
        parent.setHandleWidth(8)
        self._sizes = [320, 80]
        self.setCursor(QtCore.Qt.CursorShape.PointingHandCursor)
        self._pos_press = QtCore.QPoint()
        font = QtGui.QFont(MonospaceFontFamily, 10)
        self.setFont(font)
        # (symbol when closed, symbol when open)
        if side == "left":
            self._symbols = (">", "<")
            self._my_index = 0
        else:
            self._symbols = ("<", ">")
            self._my_index = -1

    def is_closed(self) -> bool:
        return self.splitter().sizes()[self._my_index] == 0

    def paintEvent(self, a0):
        painter = QtGui.QPainter(self)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        pen = QtGui.QPen(QtCore.Qt.GlobalColor.gray, 1.5)
        painter.setPen(pen)
        painter.setBrush(QtCore.Qt.GlobalColor.gray)
        font = painter.font()
        font.setBold(True)
        painter.setFont(font)

        width = self.width()
        height = self.height()
        if self.is_closed():
            text = self._symbols[0]
        else:
            text = self._symbols[1]
        painter.drawLine(width // 2, 3, width // 2, height // 2 - 9)
        painter.drawText(width // 2 - 2, height // 2 + 5, text)
        painter.drawLine(width // 2, height // 2 + 9, width // 2, height - 3)

    def mousePressEvent(self, a0: QtGui.QMouseEvent) -> None:
        """Collapse/expand side area."""
        self._pos_press = a0.pos()
        return super().mousePressEvent(a0)

    def mouseReleaseEvent(self, a0):
        if (a0.pos() - self._pos_press).manhattanLength() < 2:
            self.toggle()
        return super().mouseReleaseEvent(a0)

    def toggle(self):
        parent = self.splitter()
        sizes = parent.sizes()
        if self.is_closed():
            parent.setSizes(self._sizes)
        else:
            self._sizes = sizes
            sizes = [1, 1]
            sizes[self._my_index] = 0
            parent.setSizes(sizes)
        return



"""Standard plotting models."""

from himena.standards.plotting import layout, models
from himena.standards.plotting.layout import (
    BaseLayoutModel,
    SingleAxes,
    SingleStackedAxes,
    Row,
    Column,
    Grid,
    Axes,
)
from himena.standards.plotting.models import (
    Line,
    Scatter,
    Bar,
    Band,
    Span,
    ErrorBar,
    Histogram,
    Texts,
)

# 3D

from himena.standards.plotting.layout3d import (
    Axes3D,
    SingleAxes3D,
)
from himena.standards.plotting.models3d import (
    Scatter3D,
    Line3D,
    Surface3D,
    Mesh3D,
)

from himena.standards.plotting.components import StyledText, Axis, BasePlotModel
from himena.standards.plotting._api import (
    figure,
    row,
    column,
    grid,
    figure_3d,
    figure_stack,
)

__all__ = [
    "models",
    "layout",
    "figure",
    "figure_3d",
    "figure_stack",
    "row",
    "column",
    "grid",
    "BaseLayoutModel",
    "SingleAxes",
    "SingleStackedAxes",
    "Row",
    "Column",
    "Grid",
    "Axes",
    "Axis",
    "BasePlotModel",
    "Line",
    "Scatter",
    "Bar",
    "Band",
    "Span",
    "ErrorBar",
    "Histogram",
    "Texts",
    "StyledText",
    "Axes3D",
    "SingleAxes3D",
    "Scatter3D",
    "Line3D",
    "Surface3D",
    "Mesh3D",
]



from __future__ import annotations

from dataclasses import dataclass, asdict
import warnings
from himena.utils.misc import lru_cache
import json
from pathlib import Path
from cmap import Color


@dataclass(frozen=True)
class Theme:
    name: str
    background: str
    foreground: str
    base_color: str
    foreground_dim: str
    highlight_dim: str
    highlight: str
    highlight_strong: str
    background_dim: str
    background_strong: str
    inv_color: str

    @classmethod
    def from_global(cls, name: str) -> Theme:
        theme = get_global_styles().get(name, None)
        if theme is None:
            warnings.warn(
                f"Theme {name} not found. Using default theme.",
                UserWarning,
                stacklevel=2,
            )
            theme = get_global_styles()["light-purple"]
        js = asdict(theme)
        self = cls(**js)
        return self

    def format_text(self, text: str) -> str:
        for name, value in asdict(self).items():
            text = text.replace(f"#[{name}]", f"{value}")
        # replace with "light" or "dark"
        text = text.replace(
            "#[theme]", "light" if self.is_light_background() else "dark"
        )
        return text

    def is_light_background(self) -> bool:
        color = Color(self.background)
        r, g, b, a = color
        return 0.299 * r + 0.587 * g + 0.114 * b > 0.5


def _mix_colors(x: Color, y: Color, ratio: float) -> Color:
    """Mix two colors."""
    xr, xg, xb, _ = x.rgba
    yr, yg, yb, _ = y.rgba
    return Color(
        [
            xr * (1 - ratio) + yr * ratio,
            xg * (1 - ratio) + yg * ratio,
            xb * (1 - ratio) + yb * ratio,
        ]
    )


@lru_cache(maxsize=1)
def get_global_styles() -> dict[str, Theme]:
    global_styles = {}
    with open(Path(__file__).parent / "defaults.json") as f:
        js: dict = json.load(f)
        for name, style in js.items():
            bg = Color(style["background"])
            fg = Color(style["foreground"])
            base = Color(style["base_color"])
            if "foreground_dim" not in style:
                style["foreground_dim"] = _mix_colors(fg, bg, 0.6).hex
            if "background_dim" not in style:
                style["background_dim"] = _mix_colors(bg, fg, 0.1).hex
            if "background_strong" not in style:
                style["background_strong"] = _mix_colors(bg, fg, -0.1).hex
            if "highlight_dim" not in style:
                style["highlight_dim"] = _mix_colors(base, bg, 0.8).hex
            if "highlight" not in style:
                style["highlight"] = _mix_colors(base, bg, 0.6).hex
            if "highlight_strong" not in style:
                style["highlight_strong"] = _mix_colors(base, bg, 0.4).hex
            global_styles[name] = Theme(name=name, **style)
    return global_styles


def default_style() -> Theme:
    return get_global_styles()["light-green"]



import pytest
import numpy as np
from himena.standards import plotting as hplt

def test_subplots():
    row = hplt.row(2)
    row[0].plot([0, 1, 2], [3, 0, -2])
    row[1].scatter([0, 1, 2], [3, 0, -2])
    with pytest.raises(IndexError):
        row[2]

    col = hplt.column(2)
    col[0].plot([0, 1, 2], [3, 0, -2])
    col[1].scatter([0, 1, 2], [3, 0, -2])
    with pytest.raises(IndexError):
        col[2]

def test_plot_model():
    fig = hplt.figure()
    x = np.arange(5)
    fig.scatter(x, np.sin(x))
    fig.scatter(np.sin(x), edge_width=2)
    fig.plot(np.cos(x / 2), color="blue")
    fig.plot(x, np.cos(x / 2), color="blue", alpha=0.5)
    fig.bar(x, np.sin(x) / 2)
    fig.bar(np.sin(x) / 2, color="red")
    fig.bar(x, np.sin(x) / 2, color="red", edge_color="blue", edge_alpha=0.7)
    fig.errorbar(x, np.cos(x), x_error=np.full(5, 0.2), y_error=np.full(5, 0.1))
    fig.hist(np.sqrt(np.arange(100)), bins=10)
    fig.hist(np.sqrt(np.arange(100)), bins=19, orient="horizontal", stat="density")
    fig.hist(np.sqrt(np.arange(100)), bins=12, stat="probability")
    fig.band(x, np.sin(x) / 2, np.cos(x) / 2)
    fig.text([0, 1], [4, 3], ["A", "B"])
    fig.axes.title = "Title"
    fig.axes.x.lim = (0, 4)
    fig.axes.y.lim = (-1, 1)
    fig.axes.x.label = "X-axis"
    fig.axes.y.label = "Y-axis"
    fig.axes.axis_color = "red"

    # use figure properties
    fig.title = "Title"
    assert fig.title == "Title"
    fig.x.lim = (0, 5)
    assert fig.x.lim == (0, 5)
    fig.y.lim = (-1, 2)
    assert fig.y.lim == (-1, 2)
    fig.x.label = "X"
    assert fig.x.label == "X"
    fig.y.label = "Y"
    assert fig.y.label == "Y"
    fig.axis_color = "blue"
    assert fig.axis_color == "blue"



from __future__ import annotations
from qtpy import QtWidgets as QtW
import numpy as np
from numpy.typing import NDArray


def labeled(
    text: str,
    widget: QtW.QWidget,
    *more_widgets: QtW.QWidget,
    label_width: int | None = None,
) -> QtW.QWidget:
    new = QtW.QWidget()
    layout = QtW.QHBoxLayout(new)
    layout.setContentsMargins(0, 0, 0, 0)
    label = QtW.QLabel(text)
    layout.addWidget(label)
    layout.addWidget(widget)
    for w in more_widgets:
        layout.addWidget(w)
    if label_width:
        label.setFixedWidth(label_width)
    return new


def quick_min_max(
    arr: NDArray[np.number],
    down_sample_to: int = 1048576,
) -> tuple[float, float]:
    if arr.dtype.kind == "b":
        return (0.0, 1.0)
    down_sample_factor = arr.size / down_sample_to
    if down_sample_factor <= 1.0:
        return arr.min(), arr.max()
    stride = int(np.ceil(down_sample_factor))
    arr_ref = arr[::stride]
    return arr_ref.min(), arr_ref.max()



from __future__ import annotations

from typing import Iterator

from qtpy import QtWidgets as QtW
from qtpy import QtGui, QtCore

from himena.consts import MonospaceFontFamily
from himena.qt._qfinderwidget import QFinderWidget

POINT_SIZES: list[int] = [5, 6, 7, 8, 9, 10, 11, 12, 14, 16, 18, 20, 24, 28, 32, 36, 40, 48, 56, 64, 72]  # fmt: skip
TAB_SIZES: list[int] = [1, 2, 3, 4, 5, 6, 7, 8]


class QMainTextEdit(QtW.QPlainTextEdit):
    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent)
        self.setWordWrapMode(QtGui.QTextOption.WrapMode.NoWrap)
        font = QtGui.QFont(MonospaceFontFamily, 10)
        self._default_font = font
        self.setFont(font)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setTabStopDistance(self.fontMetrics().horizontalAdvance(" ") * 4)
        self._tab_size = 4
        self._highlight = None
        self._language = "Plain Text"
        self._code_theme = "default"
        self._finder_widget = None

    def is_modified(self) -> bool:
        return self.document().isModified()

    def syntax_highlight(self, lang: str | None = None):
        """Highlight syntax."""
        self._language = lang
        if lang is None or lang == "Plain Text":
            if self._highlight is not None:
                self._highlight.setDocument(None)
            return None
        from superqt.utils import CodeSyntaxHighlight

        highlight = CodeSyntaxHighlight(self.document(), lang, theme=self._code_theme)
        self._highlight = highlight
        return None

    def tab_size(self):
        return self._tab_size

    def set_tab_size(self, size: int):
        self._tab_size = size
        self.setTabStopDistance(self.fontMetrics().horizontalAdvance(" ") * size)

    def event(self, ev: QtCore.QEvent):
        try:
            if ev.type() == QtCore.QEvent.Type.KeyPress:
                assert isinstance(ev, QtGui.QKeyEvent)
                _key = ev.key()
                _mod = ev.modifiers()
                if (
                    _key == QtCore.Qt.Key.Key_Tab
                    and _mod == QtCore.Qt.KeyboardModifier.NoModifier
                ):
                    return self._tab_event()
                elif (
                    _key == QtCore.Qt.Key.Key_Tab
                    and _mod & QtCore.Qt.KeyboardModifier.ShiftModifier
                ):
                    return self._back_tab_event()
                elif _key == QtCore.Qt.Key.Key_Backtab:
                    return self._back_tab_event()
                # move selected lines up or down
                elif (
                    _key in (QtCore.Qt.Key.Key_Up, QtCore.Qt.Key.Key_Down)
                    and _mod & QtCore.Qt.KeyboardModifier.AltModifier
                ):
                    cursor = self.textCursor()
                    cursor0 = self.textCursor()
                    start = cursor.selectionStart()
                    end = cursor.selectionEnd()
                    Op = QtGui.QTextCursor.MoveOperation
                    _keep = QtGui.QTextCursor.MoveMode.KeepAnchor
                    if _key == QtCore.Qt.Key.Key_Up and min(start, end) > 0:
                        cursor0.setPosition(start)
                        cursor0.movePosition(Op.PreviousBlock)
                        cursor0.movePosition(Op.StartOfLine)
                        cursor0.movePosition(Op.EndOfLine, _keep)
                        cursor0.movePosition(Op.NextCharacter, _keep)
                        txt = cursor0.selectedText()
                        cursor0.removeSelectedText()
                        # NOTE: cursor position changed!
                        cursor0.setPosition(cursor.selectionEnd())
                        cursor0.movePosition(Op.EndOfLine)
                        if cursor0.position() == self.document().characterCount() - 1:
                            cursor0.insertText("\n")
                            txt = txt.rstrip("\u2029")
                        if cursor.position() == self.document().characterCount() - 1:
                            cursor.movePosition(Op.Up)
                        cursor0.movePosition(Op.NextCharacter)
                        cursor0.insertText(txt)
                        self.setTextCursor(cursor)
                    elif (
                        _key == QtCore.Qt.Key.Key_Down
                        and max(start, end) < self.document().characterCount() - 1
                    ):
                        cursor0.setPosition(end)
                        cursor0.movePosition(Op.EndOfLine)
                        cursor0.movePosition(Op.NextCharacter, _keep)
                        cursor0.movePosition(Op.EndOfLine, _keep)
                        txt = cursor0.selectedText()
                        cursor0.removeSelectedText()
                        # NOTE: cursor position changed!
                        cursor0.setPosition(cursor.selectionStart())
                        cursor0.movePosition(Op.StartOfLine)
                        if cursor0.position() == 0:
                            cursor0.insertText("\n")
                            txt = txt.lstrip("\u2029")
                        cursor0.movePosition(Op.PreviousCharacter)
                        cursor0.insertText(txt)
                        self.setTextCursor(cursor)
                    return True
                elif _key in (QtCore.Qt.Key.Key_Enter, QtCore.Qt.Key.Key_Return):
                    # get current line, check if it has tabs at the beginning
                    # if yes, insert the same number of tabs at the next line
                    self._new_line_event()
                    return True
                elif (
                    _key == QtCore.Qt.Key.Key_Backspace
                    and _mod == QtCore.Qt.KeyboardModifier.NoModifier
                ):
                    # delete 4 spaces
                    _cursor = self.textCursor()
                    _cursor.movePosition(
                        QtGui.QTextCursor.MoveOperation.StartOfLine,
                        QtGui.QTextCursor.MoveMode.KeepAnchor,
                    )
                    line = _cursor.selectedText()
                    if line.endswith("    ") and not self.textCursor().hasSelection():
                        for _ in range(4):
                            self.textCursor().deletePreviousChar()
                        return True
                elif (
                    _key == QtCore.Qt.Key.Key_D
                    and _mod & QtCore.Qt.KeyboardModifier.ControlModifier
                ):
                    return self._select_word_event()
                elif (
                    _key == QtCore.Qt.Key.Key_L
                    and _mod & QtCore.Qt.KeyboardModifier.ControlModifier
                ):
                    return self._select_line_event()
                elif (
                    _key == QtCore.Qt.Key.Key_Home
                    and _mod == QtCore.Qt.KeyboardModifier.NoModifier
                ):
                    return self._home_event()
                elif (
                    _key == QtCore.Qt.Key.Key_V
                    and _mod & QtCore.Qt.KeyboardModifier.ControlModifier
                ):
                    clip = QtGui.QGuiApplication.clipboard()
                    text = clip.text().replace("\n\t", "\n" + " " * self.tab_size())
                    cursor = self.textCursor()
                    cursor.insertText(text)
                    return True
                elif (
                    _key == QtCore.Qt.Key.Key_Comma
                    and _mod & QtCore.Qt.KeyboardModifier.ControlModifier
                    and _mod & QtCore.Qt.KeyboardModifier.ShiftModifier
                ) or (
                    _key == QtCore.Qt.Key.Key_Greater
                    and _mod & QtCore.Qt.KeyboardModifier.ControlModifier
                ):
                    self.setFont(change_point_size(self.font(), 1))
                    return True
                elif (
                    _key == QtCore.Qt.Key.Key_Period
                    and _mod & QtCore.Qt.KeyboardModifier.ControlModifier
                    and _mod & QtCore.Qt.KeyboardModifier.ShiftModifier
                ) or (
                    _key == QtCore.Qt.Key.Key_Less
                    and _mod & QtCore.Qt.KeyboardModifier.ControlModifier
                ):
                    self.setFont(change_point_size(self.font(), -1))
                    return True
                elif (
                    _key == QtCore.Qt.Key.Key_0
                    and _mod & QtCore.Qt.KeyboardModifier.ControlModifier
                ):
                    self.setFont(self._default_font)
                    return True

        except Exception:
            pass
        return super().event(ev)

    def _iter_selected_lines(self) -> Iterator[QtGui.QTextCursor]:
        """Iterate text cursors for each selected line."""
        _cursor = self.textCursor()
        start, end = sorted([_cursor.selectionStart(), _cursor.selectionEnd()])
        _cursor.setPosition(start)
        _cursor.movePosition(QtGui.QTextCursor.MoveOperation.StartOfLine)
        nline = 0
        while True:
            _cursor.movePosition(QtGui.QTextCursor.MoveOperation.EndOfLine)
            _cursor.movePosition(QtGui.QTextCursor.MoveOperation.NextCharacter)
            nline += 1
            if _cursor.position() >= end:
                break

        _cursor.setPosition(start)
        for _ in range(nline):
            _cursor.movePosition(QtGui.QTextCursor.MoveOperation.EndOfLine)
            yield _cursor
            _cursor.movePosition(QtGui.QTextCursor.MoveOperation.EndOfLine)
            _cursor.movePosition(QtGui.QTextCursor.MoveOperation.NextCharacter)

    def _tab_event(self):
        if self.textCursor().hasSelection():
            for cursor in self._iter_selected_lines():
                self._add_at_the_start(" " * self.tab_size(), cursor)
        else:
            line = self._text_of_line_before_cursor()
            nspace = line.count(" ")
            if nspace % 4 == 0:
                self.textCursor().insertText(" " * self.tab_size())
            else:
                self.textCursor().insertText(" " * 4 - nspace % 4)
        return True

    def _add_at_the_start(self, text: str, cursor: QtGui.QTextCursor):
        cursor.movePosition(QtGui.QTextCursor.MoveOperation.StartOfLine)
        cursor.insertText(text)

    def _remove_at_the_start(self, text: str, cursor: QtGui.QTextCursor):
        line = cursor.block().text()
        if line.startswith(text):
            cursor.movePosition(QtGui.QTextCursor.MoveOperation.StartOfLine)
            cursor.movePosition(
                QtGui.QTextCursor.MoveOperation.Right,
                QtGui.QTextCursor.MoveMode.KeepAnchor,
                len(text),
            )
            cursor.removeSelectedText()

    def _back_tab_event(self):
        # unindent
        for cursor in self._iter_selected_lines():
            self._remove_at_the_start(" " * self.tab_size(), cursor)
        return True

    def _text_of_line_before_cursor(self):
        _cursor = self.textCursor()
        _cursor.movePosition(
            QtGui.QTextCursor.MoveOperation.StartOfLine,
            QtGui.QTextCursor.MoveMode.KeepAnchor,
        )
        return _cursor.selectedText()

    def _select_word_event(self):
        cursor = self.textCursor()
        cursor.movePosition(QtGui.QTextCursor.MoveOperation.NextWord)
        cursor.movePosition(
            QtGui.QTextCursor.MoveOperation.PreviousWord,
            QtGui.QTextCursor.MoveMode.KeepAnchor,
        )
        self.setTextCursor(cursor)
        return True

    def _select_line_event(self):
        cursor = self.textCursor()
        cursor.movePosition(QtGui.QTextCursor.MoveOperation.StartOfLine)
        cursor.movePosition(
            QtGui.QTextCursor.MoveOperation.EndOfLine,
            QtGui.QTextCursor.MoveMode.KeepAnchor,
        )
        cursor.movePosition(
            QtGui.QTextCursor.MoveOperation.NextCharacter,
            QtGui.QTextCursor.MoveMode.KeepAnchor,
        )
        self.setTextCursor(cursor)
        return True

    def _home_event(self):
        # fn + left
        cursor = self.textCursor()
        cursor.movePosition(
            QtGui.QTextCursor.MoveOperation.StartOfLine,
            QtGui.QTextCursor.MoveMode.KeepAnchor,
        )
        text = cursor.selectedText()
        if all(c == " " for c in text):
            cursor.clearSelection()
        else:
            text_lstrip = text.lstrip()
            nmove = len(text) - len(text_lstrip)
            cursor.clearSelection()
            for _ in range(nmove):
                cursor.movePosition(QtGui.QTextCursor.MoveOperation.Right)

        self.setTextCursor(cursor)
        return True

    def _new_line_event(self):
        line = self._text_of_line_before_cursor()
        cursor = self.textCursor()
        line_rstripped = line.rstrip()
        indent = _get_indents(line, self.tab_size())
        if line_rstripped == "":
            cursor.insertText("\n" + indent)
            self.setTextCursor(cursor)
            return
        cursor.insertText("\n" + indent)
        self.setTextCursor(cursor)

    def _find_string(self):
        if self._finder_widget is None:
            self._finder_widget = QFinderWidget(self)
        self._finder_widget.show()
        self._align_finder()

    def resizeEvent(self, event):
        if self._finder_widget is not None:
            self._align_finder()
        super().resizeEvent(event)

    def _align_finder(self):
        if fd := self._finder_widget:
            vbar = self.verticalScrollBar()
            if vbar.isVisible():
                fd.move(self.width() - fd.width() - vbar.width() - 3, 5)
            else:
                fd.move(self.width() - fd.width() - 3, 5)


def _get_indents(text: str, tab_spaces: int = 4) -> str:
    chars = []
    for c in text:
        if c == " ":
            chars.append(" ")
        elif c == "\t":
            chars.append(" " * tab_spaces)
        else:
            break
    return "".join(chars)


def change_point_size(cur_font: QtGui.QFont, step: int) -> QtGui.QFont:
    current_size = cur_font.pointSize()
    nmax = len(POINT_SIZES)
    cur_idx = nmax - 1
    for idx, size in enumerate(POINT_SIZES):
        if current_size <= size:
            cur_idx = idx
            break
    next_idx = min(max(cur_idx + step, 0), nmax - 1)
    new_size = POINT_SIZES[next_idx]
    cur_font.setPointSize(new_size)
    return cur_font



from __future__ import annotations
from pathlib import Path

from qtpy import QtWidgets as QtW, QtCore

from himena.consts import StandardType
from himena.types import WidgetDataModel
from himena.plugins import validate_protocol
from himena.qt._utils import get_main_window


class QReaderNotFound(QtW.QWidget):
    """No reader function is defined for the file.

    The content of the file is not read yet. This widget only contains the file path.
    "Open as text" button does not guarantee that the file can be opened as text and
    does not check the file size. Please make sure the file is actually a text file by
    yourself.

    Remote files are also supported.
    """

    __himena_widget_id__ = "builtins:QReaderNotFound"

    def __init__(self):
        super().__init__()
        layout = QtW.QVBoxLayout(self)
        self._label = QtW.QLabel("Reader not found.")
        self._label.setTextInteractionFlags(
            QtCore.Qt.TextInteractionFlag.TextSelectableByMouse
        )
        self._label.setWordWrap(True)
        self._open_as_text_button = QtW.QPushButton("Open as text")
        self._open_as_text_button.clicked.connect(self._open_as_text)
        self._file_path: Path | None = None

        layout.addWidget(self._label)
        layout.addWidget(self._open_as_text_button)

    @validate_protocol
    def update_model(self, model: WidgetDataModel[Path]):
        self._file_path = model.value
        if not self._file_path.exists():
            _byte = -1
        else:
            _byte = self._file_path.stat().st_size
        if _byte < 0:
            _size = "???"
        elif _byte < 1024:
            _size = f"{_byte} B"
        elif _byte < 1024**2:
            _size = f"{_byte / 1024:.2f} KB"
        elif _byte < 1024**3:
            _size = f"{_byte / 1024**2:.2f} MB"
        else:
            _size = f"{_byte / 1024**3:.2f} GB"
        self._label.setText(f"Reader not found for {model.value.name} ({_size})")

    @validate_protocol
    def to_model(self) -> WidgetDataModel[Path]:
        return WidgetDataModel(
            value=self._file_path,
            type=self.model_type(),
        )

    @validate_protocol
    def model_type(self) -> str:
        return StandardType.READER_NOT_FOUND

    def _open_as_text(self):
        get_main_window(self).exec_action("builtins:open-as-text-anyway")



from __future__ import annotations
from typing import Literal

from qtpy import QtWidgets as QtW, QtCore, QtGui
from magicgui import widgets as mgw
from himena.standards import model_meta
from himena.qt._utils import qsignal_blocker
from himena.consts import MonospaceFontFamily


class QDimsSlider(QtW.QWidget):
    """Dimension sliders for an array."""

    valueChanged = QtCore.Signal(tuple)

    def __init__(self):
        super().__init__()
        self._sliders: list[_QAxisSlider] = []
        layout = QtW.QVBoxLayout(self)
        layout.setContentsMargins(2, 2, 2, 2)
        layout.setSpacing(2)
        self._yx_axes: list[model_meta.DimAxis] = [
            model_meta.DimAxis(name="y"),
            model_meta.DimAxis(name="x"),
        ]

    def count(self) -> int:
        """Number of sliders."""
        return len(self._sliders)

    def maximums(self) -> tuple[int, ...]:
        """Return the maximum values of the sliders."""
        return tuple(slider._slider.maximum() for slider in self._sliders)

    def set_dimensions(
        self,
        shape: tuple[int, ...],
        axes: list[model_meta.DimAxis] | None = None,
        is_rgb: bool = False,
    ):
        ndim = len(shape)
        ndim_rem = ndim - 3 if is_rgb else ndim - 2
        nsliders = len(self._sliders)
        if nsliders > ndim_rem:
            for i in range(ndim_rem, nsliders):
                slider = self._sliders.pop()
                self.layout().removeWidget(slider)
                slider.deleteLater()
        elif nsliders < ndim_rem:
            for i in range(nsliders, ndim_rem):
                self._make_slider(shape[i])
        # update axis names
        if axes is None:
            axes = [model_meta.DimAxis(name=f"axis {i}") for i in range(ndim_rem)]
        _axis_width_max = 0
        _index_width_max = 0
        for axis, slider in zip(axes, self._sliders):
            aname = axis.name
            slider.update_from_axis(axis)
            # TODO: show scale, unit and origin
            width = slider._name_label.fontMetrics().boundingRect(aname).width()
            _axis_width_max = max(_axis_width_max, width)
            _i_max = slider._slider.maximum()
            width = (
                slider._index_label.fontMetrics()
                .boundingRect(f"{_i_max}/{_i_max}")
                .width()
            )
            _index_width_max = max(_index_width_max, width)
        for slider in self._sliders:
            slider._name_label.setFixedWidth(_axis_width_max + 6)
            slider._index_label.setFixedWidth(_index_width_max + 6)
        self._yx_axes = axes[-2:]

    def set_play_setting(self, setting: model_meta.ImagePlaySetting) -> None:
        """Set play setting for the last slider (the first non-yx axis)."""
        if not self._sliders:
            return
        slider = self._sliders[-1]
        slider._play_timer.setInterval(int(setting.interval * 1000))
        slider._play_back_mode = setting.mode

    def to_dim_axes(self) -> list[model_meta.DimAxis]:
        axes = [slider.to_axis() for slider in self._sliders]
        axes.extend(self._yx_axes)
        return axes

    def to_play_setting(self) -> model_meta.ImagePlaySetting | None:
        for slider in self._sliders:
            if slider.to_axis().name in ("t", "time", "frame"):
                return model_meta.ImagePlaySetting(
                    interval=slider._play_timer.interval() / 1000,
                    mode=slider._play_back_mode,
                )

    def _make_slider(self, size: int) -> _QAxisSlider:
        slider = _QAxisSlider()
        self._sliders.append(slider)
        self.layout().addWidget(slider, alignment=QtCore.Qt.AlignmentFlag.AlignBottom)
        slider.setRange(0, size - 1)
        slider._slider.valueChanged.connect(self._emit_value)
        return slider

    def _emit_value(self):
        self.valueChanged.emit(self.value())

    def value(self) -> tuple[int, ...]:
        return tuple(slider._slider.value() for slider in self._sliders)

    def setValue(self, value: tuple[int, ...]) -> None:
        self.set_value_no_emit(value)
        self.valueChanged.emit(value)

    def set_value_no_emit(self, value: tuple[int, ...]) -> None:
        if len(value) != len(self._sliders):
            raise ValueError(f"Expected {len(self._sliders)} values, got {len(value)}")
        for slider, val in zip(self._sliders, value):
            if val == -1:  # flattened axis, no need to update slider
                continue
            with qsignal_blocker(slider):
                slider._slider.setValue(val)

    def axis_names(self) -> list[str]:
        return [slider._name_label.text() for slider in self._sliders]

    def set_axis_names(self, names: list[str]) -> None:
        for slider, name in zip(self._sliders, names):
            slider._name_label.setText(name)


class _QSliderButton(QtW.QPushButton):
    """A button for the slider."""

    def __init__(self, text: str):
        super().__init__(text)
        self.setStyleSheet("_QSliderButton { padding: 0px; margin: 0px}")
        self.setFixedSize(12, 12)
        font = QtGui.QFont(MonospaceFontFamily, 8)
        self.setFont(font)
        self.setAutoRepeat(True)
        self.setAutoDefault(False)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)


class _QAxisSlider(QtW.QWidget):
    """A slider widget for an axis."""

    def __init__(self) -> None:
        super().__init__()
        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self._name_label = QtW.QLabel()
        self._name_label.setFixedWidth(60)
        self._name_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)

        self._play_btn = _QSliderButton("▶")
        self._play_btn.setToolTip("Play along this axis (right click to configure)")
        self._play_btn.setCheckable(True)
        self._play_btn.setAutoRepeat(False)
        self._play_btn.setContextMenuPolicy(
            QtCore.Qt.ContextMenuPolicy.CustomContextMenu
        )
        self._play_btn.customContextMenuRequested.connect(
            lambda pos: self._make_context_menu_for_play_btn().exec(
                self._play_btn.mapToGlobal(pos)
            )
        )
        self._play_timer = QtCore.QTimer(self)
        self._play_timer.setInterval(100)
        self._play_timer.timeout.connect(self._on_play_timer_timeout)
        self._play_back_mode: Literal["once", "loop", "pingpong"] = "loop"
        self._play_increment = 1
        self._play_btn.clicked.connect(self._on_play_clicked)

        self._prev_btn = _QSliderButton("<")
        self._prev_btn.clicked.connect(self._on_prev_clicked)

        self._slider = QtW.QScrollBar(QtCore.Qt.Orientation.Horizontal)
        self._slider.setContentsMargins(0, 0, 0, 0)
        self._slider.setSingleStep(1)
        self._slider.setPageStep(1)
        self._slider.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Fixed
        )
        self._next_btn = _QSliderButton(">")
        self._next_btn.clicked.connect(self._on_next_clicked)
        self._index_label = QtW.QLabel()
        self._index_label.setCursor(QtCore.Qt.CursorShape.IBeamCursor)
        self._index_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        self._slider.valueChanged.connect(self._on_slider_changed)

        layout.addWidget(self._name_label)
        layout.addWidget(self._play_btn)
        layout.addWidget(self._prev_btn)
        layout.addWidget(self._slider)
        layout.addWidget(self._next_btn)
        layout.addWidget(
            self._index_label, alignment=QtCore.Qt.AlignmentFlag.AlignRight
        )
        self._axis = model_meta.DimAxis(name="")
        self._edit_value_line = QCurrentIndexEdit(self)
        self._edit_value_line.setFont(self._index_label.font())
        self._edit_value_line.edited.connect(self._on_edit_finished)

    def update_from_axis(self, axis: model_meta.DimAxis):
        self._name_label.setText(axis.name)
        self._axis = axis.model_copy()

    def to_axis(self) -> model_meta.DimAxis:
        return self._axis

    def text(self) -> str:
        return self._name_label.text()

    def setRange(self, start: int, end: int) -> None:
        self._slider.setRange(start, end)
        self._index_label.setText(f"{self._slider.value()}/{end}")
        self._edit_value_line.setValidator(
            QtGui.QIntValidator(start, end, self._edit_value_line)
        )

    def _on_slider_changed(self, value: int) -> None:
        self._index_label.setText(f"{value}/{self._slider.maximum()}")

    def _on_edit_finished(self, value: int) -> None:
        self._slider.setValue(value)
        self._index_label.setText(f"{value}/{self._slider.maximum()}")

    def mouseDoubleClickEvent(self, a0):
        if self._index_label.geometry().contains(a0.pos()):
            self._edit_value_line._double_clicked(self._index_label)
        else:
            return super().mouseDoubleClickEvent(a0)

    def increment_value(self, val: int):
        self._stop_play()
        new_val = self._slider.value() + val
        self._slider.setValue(min(max(new_val, 0), self._slider.maximum()))

    def _on_prev_clicked(self):
        self.increment_value(-1)

    def _on_next_clicked(self):
        self.increment_value(1)

    def _on_play_timer_timeout(self):
        if self._play_btn.isChecked():
            new_value = self._slider.value() + self._play_increment
            if new_value > self._slider.maximum():
                if self._play_back_mode == "loop":
                    new_value = 0
                elif self._play_back_mode == "pingpong":
                    new_value = self._slider.maximum() - 1
                    self._play_increment = -self._play_increment
                else:
                    self._stop_play()
            elif new_value < 0:
                if self._play_back_mode == "loop":
                    new_value = self._slider.maximum()
                elif self._play_back_mode == "pingpong":
                    new_value = 1
                    self._play_increment = -self._play_increment
                else:
                    self._stop_play()
            self._slider.setValue(new_value)

    def _stop_play(self):
        self._play_btn.setChecked(False)
        self._play_btn.setText("▶")
        self._play_timer.stop()

    def _on_play_clicked(self):
        if self._play_btn.isChecked():
            if (
                self._play_back_mode == "once"
                and self._slider.value() == self._slider.maximum()
            ):
                self._slider.setValue(0)
            self._play_timer.start()
            self._play_btn.setText("■")
        else:
            self._play_timer.stop()
            self._play_btn.setText("▶")

    def _make_context_menu_for_play_btn(self) -> QtW.QMenu:
        menu = QtW.QMenu(self)
        direction = mgw.ComboBox(
            label="Direction",
            choices=[("forward", 1), ("backward", -1)],
            value=1,
        )
        fps = mgw.FloatSpinBox(
            label="FPS",
            min=0.1,
            max=100.0,
            value=1000 / self._play_timer.interval(),
            step=10,
        )
        loop = mgw.ComboBox(
            label="Mode",
            choices=["once", "loop", "pingpong"],
            value=self._play_back_mode,
        )
        container = mgw.Container(widgets=[direction, fps, loop])
        fps.changed.connect(lambda v: self._play_timer.setInterval(int(1000 / v)))
        loop.changed.connect_setattr(self, "_play_back_mode", maxargs=1)
        direction.changed.connect_setattr(self, "_play_increment", maxargs=1)
        widget_action = QtW.QWidgetAction(menu)
        widget_action.setDefaultWidget(container.native)
        menu.addAction(widget_action)
        return menu


class QCurrentIndexEdit(QtW.QLineEdit):
    """A line edit for current index."""

    edited = QtCore.Signal(int)

    def __init__(self, parent: QtW.QWidget):
        super().__init__()
        self.setParent(
            parent,
            QtCore.Qt.WindowType.Dialog | QtCore.Qt.WindowType.FramelessWindowHint,
        )
        self.hide()
        self.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)

    def _finish_edit(self):
        self.edited.emit(int(self.text()))
        self._cancel_edit()

    def _cancel_edit(self):
        self.hide()
        self.parentWidget().setFocus()

    def _double_clicked(self, label: QtW.QLabel):
        """Start editing the current index."""
        self.show()
        current, _max = label.text().split("/", 1)
        dx = self.fontMetrics().boundingRect(f"/{_max}").width() + 1
        size = label.size()
        self.resize(size.width() - dx, size.height())
        geo = label.geometry()
        self.move(self.parentWidget().mapToGlobal(geo.topLeft()))
        self.setText(current)
        self.setFocus()
        self.selectAll()

    def focusOutEvent(self, a0):
        self._finish_edit()
        return super().focusOutEvent(a0)

    def keyPressEvent(self, a0):
        if a0.key() == QtCore.Qt.Key.Key_Return:
            self._finish_edit()
        elif a0.key() == QtCore.Qt.Key.Key_Escape:
            self._cancel_edit()
        else:
            return super().keyPressEvent(a0)



"""Standard ROI (Region of Interest) classes for images."""

from himena.standards.roi._base import (
    RoiModel,
    Roi1D,
    Roi2D,
    Roi3D,
    default_roi_label,
    pick_roi_model,
)

from himena.standards.roi.core import (
    SpanRoi,
    PointRoi1D,
    PointsRoi1D,
    RectangleRoi,
    RotatedRectangleRoi,
    EllipseRoi,
    RotatedEllipseRoi,
    PointsRoi2D,
    PointRoi2D,
    SegmentedLineRoi,
    PolygonRoi,
    CircleRoi,
    LineRoi,
    SplineRoi,
)
from himena.standards.roi._list import RoiListModel

__all__ = [
    "RoiModel",
    "Roi1D",
    "Roi2D",
    "Roi3D",
    "SpanRoi",
    "PointRoi1D",
    "PointsRoi1D",
    "RectangleRoi",
    "RotatedRectangleRoi",
    "EllipseRoi",
    "RotatedEllipseRoi",
    "PointsRoi2D",
    "PointRoi2D",
    "SegmentedLineRoi",
    "PolygonRoi",
    "CircleRoi",
    "LineRoi",
    "SplineRoi",
    "RoiListModel",
    "default_roi_label",
    "pick_roi_model",
]



from dataclasses import dataclass
from typing import Any

from himena.consts import DefaultFontFamily
from himena.plugins import config_field


@dataclass
class MatplotlibCanvasConfigs:
    """Matplotlib canvas configurations."""

    font_size: int = config_field(10, tooltip="Default font size", label="font.size")
    font_family: str = config_field(
        DefaultFontFamily, tooltip="Default font family", label="font.family"
    )
    axes_spines_top: bool = config_field(
        False, tooltip="Show top spine", label="axes.spines.top"
    )
    axes_spines_right: bool = config_field(
        False, tooltip="Show right spine", label="axes.spines.right"
    )

    def to_dict(self) -> dict[str, Any]:
        return {
            "font.size": self.font_size,
            "font.family": self.font_family,
            "axes.spines.top": self.axes_spines_top,
            "axes.spines.right": self.axes_spines_right,
        }



from __future__ import annotations

from qtpy import QtWidgets as QtW
from qtpy.QtCore import Qt, Signal
from himena_builtins.qt.widgets._table_components._base import QTableBase
from himena.qt._qlineedit import QIntLineEdit
from himena.qt._utils import qsignal_blocker


class QSelectionRangeEdit(QtW.QGroupBox):
    sliceChanged = Signal(object)

    def __init__(
        self,
        table: QTableBase | None = None,
        parent: QtW.QWidget | None = None,
    ):
        super().__init__(parent)
        self.setStyleSheet("QSelectionRangeEdit {margin: 0px;}")
        self._qtable: QTableBase | None = None
        self.setLayout(QtW.QHBoxLayout())
        self.setContentsMargins(0, 0, 0, 0)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setAlignment(Qt.AlignmentFlag.AlignRight)
        inner = QtW.QWidget()
        self.layout().addWidget(inner)
        layout = QtW.QHBoxLayout()
        layout.setContentsMargins(8, 0, 8, 0)
        layout.setSpacing(0)
        inner.setLayout(layout)
        self._r_start = _int_line_edit("Row starting index", self._rstart_changed)
        self._r_stop = _int_line_edit("Row stopping index", self._rstop_changed)
        self._c_start = _int_line_edit("Column starting index", self._cstart_changed)
        self._c_stop = _int_line_edit("Column stopping index", self._cstop_changed)
        self._r_colon = _label(":")
        self._c_colon = _label(":")
        rbox = _hbox(self._r_start, self._r_colon, self._r_stop)
        cbox = _hbox(self._c_start, self._c_colon, self._c_stop)

        layout.addWidget(QtW.QLabel("Select ("))
        layout.addWidget(rbox)
        layout.addWidget(_label(", "))
        layout.addWidget(cbox)
        layout.addWidget(QtW.QLabel(")"))

        if table is not None:
            self.connect_table(table)
        self.sliceChanged.connect(self._slice_changed)
        self.setMaximumWidth(150)

    def connect_table(self, table: QTableBase):
        if self._qtable is not None:
            self._qtable._selection_model.moved.disconnect(self._selection_changed)
        self._qtable = table
        self._qtable._selection_model.moved.connect(self._selection_changed)
        self.setSlice(((0, 1), (0, 1)))
        self._selection_changed()

    def _int_gt(self, s: str, default: int) -> int:
        if s.strip() == "":
            return default
        return max(int(s), default)

    def _int_lt(self, s: str, default: int) -> int:
        if s.strip() == "":
            return default
        return min(int(s), default)

    def _selection_changed(self):
        if self._qtable is None:
            return
        sels = self._qtable._selection_model.ranges
        if len(sels) == 0:
            return
        rsl, csl = sels[-1]
        self.setSlice(((rsl.start, rsl.stop), (csl.start, csl.stop)))

    def _slice_changed(self, sl: tuple[tuple[int, int], tuple[int, int]]):
        if self._qtable is None:
            return
        rsl, csl = sl
        self._qtable._selection_model.set_ranges([(slice(*rsl), slice(*csl))])
        idx = rsl[0] - 1, csl[0] - 1
        self._qtable._selection_model.current_index = idx
        index = self._qtable.model().index(*idx)
        self._qtable.setCurrentIndex(index)
        return None

    def slice(self) -> tuple[tuple[int, int], tuple[int, int]]:
        if self._qtable is None:
            return ((0, 1), (0, 1))
        rsl = (
            self._int_gt(self._r_start.text(), 0),
            self._int_lt(self._r_stop.text(), self._qtable.model().rowCount()),
        )
        csl = (
            self._int_gt(self._c_start.text(), 0),
            self._int_lt(self._c_stop.text(), self._qtable.model().columnCount()),
        )
        return rsl, csl

    def setSlice(self, sl: tuple[tuple[int, int], tuple[int, int]]):
        rsl, csl = sl
        rstart, rstop = rsl
        cstart, cstop = csl
        with qsignal_blocker(self):
            self._r_start.setText(str(rstart))
            self._r_stop.setText(str(rstop))
            self._c_start.setText(str(cstart))
            self._c_stop.setText(str(cstop))
        if rstart is not None and rstop is not None and rstop == rstart + 1:
            self._r_stop.hide()
            self._r_colon.hide()
        else:
            self._r_stop.show()
            self._r_colon.show()
        if cstart is not None and cstop is not None and cstop == cstart + 1:
            self._c_stop.hide()
            self._c_colon.hide()
        else:
            self._c_stop.show()
            self._c_colon.show()
        return None

    def _rstart_changed(self, txt: str):
        rstop = self._r_stop.text()
        if txt and rstop:
            if not self._r_stop.isVisible() or int(rstop) <= int(txt):
                self._r_stop.setText(str(int(txt) + 1))
            return self.sliceChanged.emit(self.slice())

    def _rstop_changed(self, txt: str):
        rstart = self._r_start.text()
        if txt and rstart and int(rstart) >= int(txt):
            int_rstop = int(txt)
            if int_rstop > 1:
                self._r_start.setText(str(int_rstop - 1))
            else:
                self._r_start.setText("0")
                self._r_stop.setText("1")
        return self.sliceChanged.emit(self.slice())

    def _cstart_changed(self, txt: str):
        cstop = self._c_stop.text()
        if txt and cstop:
            if not self._c_stop.isVisible() or int(cstop) <= int(txt):
                self._c_stop.setText(str(int(txt) + 1))
            return self.sliceChanged.emit(self.slice())

    def _cstop_changed(self, txt: str):
        cstart = self._c_start.text()
        if txt and cstart and int(cstart) >= int(txt):
            int_cstop = int(txt)
            if int_cstop > 1:
                self._c_start.setText(str(int_cstop - 1))
            else:
                self._c_start.setText("0")
                self._c_stop.setText("1")
        return self.sliceChanged.emit(self.slice())


def _label(text: str):
    label = QtW.QLabel(text)
    label.setAlignment(Qt.AlignmentFlag.AlignCenter)
    label.setFixedWidth(8)
    return label


def _int_line_edit(tooltip: str, text_changed_callback) -> QIntLineEdit:
    out = QIntLineEdit()
    out.setObjectName("TableIndex")
    out.setToolTip(tooltip)
    out.setAlignment(Qt.AlignmentFlag.AlignRight)
    out.textChanged.connect(text_changed_callback)
    return out


def _hbox(*widgets: QtW.QWidget) -> QtW.QWidget:
    box = QtW.QWidget()
    layout = QtW.QHBoxLayout(box)
    layout.setContentsMargins(0, 0, 0, 0)
    layout.setSpacing(0)
    for widget in widgets:
        layout.addWidget(widget)
    box.setFixedWidth(42)
    return box



from __future__ import annotations
from typing import Any, Generic, TypeVar, overload

from himena.plugins import _checker
from himena.types import DragDataModel, DropResult, Size, WidgetDataModel
from himena.style import default_style
from himena.widgets import current_instance
from himena.workflow import ProgrammaticMethod

_W = TypeVar("_W")


class WidgetTester(Generic[_W]):
    def __init__(self, widget: _W):
        self._widget = widget

    def __enter__(self) -> WidgetTester[_W]:
        _checker.call_theme_changed_callback(self._widget, default_style())
        _checker.call_widget_activated_callback(self._widget)
        _checker.call_widget_added_callback(self._widget)
        if hasattr(self._widget, "control_widget"):
            self._widget.control_widget()
        if hasattr(self._widget, "size_hint"):
            hint = self._widget.size_hint()
            _checker.call_widget_resized_callback(
                self._widget, Size(200, 200), Size(*hint)
            )
        if hasattr(self._widget, "is_editable"):
            self._widget.is_editable()
        if hasattr(self._widget, "set_editable"):
            self._widget.set_editable(False)
            self._widget.set_editable(True)
        return self

    def __exit__(self, *args):
        if hasattr(self._widget, "is_modified"):
            self._widget.is_modified()
        _checker.call_widget_closed_callback(self._widget)

    @overload
    def update_model(self, model: WidgetDataModel) -> WidgetTester[_W]: ...
    @overload
    def update_model(
        self,
        value: Any,
        *,
        type: str | None = None,
        metadata: Any | None = None,
        **kwargs,
    ) -> WidgetTester[_W]: ...

    def update_model(self, value, **kwargs) -> WidgetTester[_W]:
        model = self._norm_model_input(value, **kwargs)
        self._widget.update_model(model)
        return self

    def to_model(self) -> WidgetDataModel:
        model = self._widget.to_model()
        if not isinstance(model, WidgetDataModel):
            raise TypeError(
                f"Widget {self.widget!r} returned {model!r}, expected WidgetDataModel"
            )
        model.workflow = ProgrammaticMethod().construct_workflow()
        return model

    def cycle_model(self) -> tuple[WidgetDataModel, WidgetDataModel]:
        """Cycle `update_model` and `to_model` and return both.

        This function is useful for testing the consistency of widget's `update_model`
        and `to_model`.
        """
        model = self.to_model()
        self.update_model(model)
        return model, self.to_model()

    @overload
    def drop_model(self, model: WidgetDataModel) -> DropResult: ...
    @overload
    def drop_model(
        self,
        value: Any,
        *,
        type: str | None = None,
        metadata: Any | None = None,
        **kwargs,
    ) -> DropResult: ...

    def drop_model(self, value, **kwargs):
        """Emulate dropping a model into the widget.

        The input can be either a `WidgetDataModel` or the input of `WidgetDataModel`.

        ``` python
        tester.drop_model(WidgetDataModel(value=..., type=...))
        tester.drop_model(value=..., type=...)
        ```
        """
        model = self._norm_model_input(value, **kwargs)
        drag_data_model = DragDataModel(getter=model, type=model.type)
        if not drag_data_model.widget_accepts_me(self.widget):
            try:
                allowed = self.widget.allowed_drop_types()
            except AttributeError:
                allowed = []
            raise ValueError(
                f"Widget {self.widget!r} does not accept dropping {model.type}. Allowed types are: {allowed}"
            )
        model.workflow = ProgrammaticMethod().construct_workflow()
        out = self.widget.dropped_callback(model)
        if out is None:
            out = DropResult()
        elif not isinstance(out, DropResult):
            raise TypeError(
                f"Widget {self.widget!r} returned {out!r}, expected DropResult"
            )
        if out.command_id is not None:
            returned_model = current_instance().exec_action(
                out.command_id,
                model_context=self.to_model(),
                with_params=out.with_params,
                process_model_output=False,
            )
            if not isinstance(returned_model, WidgetDataModel):
                raise ValueError(
                    f"Command {out.command_id} did not return a WidgetDataModel."
                )
            self.update_model(returned_model)
        return out

    def is_modified(self) -> bool:
        return self._widget.is_modified()

    @property
    def widget(self) -> _W:
        return self._widget

    def _norm_model_input(self, val, **kwargs) -> WidgetDataModel:
        if isinstance(val, WidgetDataModel):
            if kwargs:
                raise TypeError("Cannot specify both model and kwargs")
            return val
        else:
            if kwargs.get("type") is None:
                try:
                    kwargs["type"] = self._widget.model_type()
                except AttributeError:
                    raise TypeError("`type` argument must be specified") from None
            return WidgetDataModel(value=val, **kwargs)



from __future__ import annotations
from typing import TYPE_CHECKING
import weakref
from app_model.types import KeyCode

if TYPE_CHECKING:
    from himena.widgets import MainWindow


class KeySet:
    def __init__(self, ui: MainWindow):
        self._ui_ref = weakref.ref(ui)

    def contains(self, key: int | str | KeyCode) -> bool:
        """Check if the given key is currently pressed."""
        if isinstance(key, str):
            _key = KeyCode.from_string(key)
        elif isinstance(key, KeyCode):
            _key = key
        elif isinstance(key, int):
            _key = KeyCode(key)
        else:
            raise TypeError(f"key must be int, str, or KeyCode, not {type(key)}")
        ui = self._ui_ref()
        if ui is None:
            return False

        return _key.value in ui._backend_main_window._keys_as_set()

    __contains__ = contains

    def __repr__(self) -> str:
        ui = self._ui_ref()
        if ui is None:
            return "KeySet(<deleted>)"
        keys = ui._backend_main_window._keys_as_set()
        strs = " ".join(KeyCode(k).os_symbol() for k in keys)
        return f"KeySet({strs})"



from __future__ import annotations

from typing import Generic, TypeVar

_E = TypeVar("_E")


class MouseEventHandler(Generic[_E]):
    """General mouse event handler class."""

    def pressed(self, event: _E):
        """Callback for when the mouse is pressed."""

    def moved(self, event: _E):
        """Callback for when the mouse is moved."""

    def released(self, event: _E):
        """Callback for when the mouse is released."""

    def double_clicked(self, event: _E):
        """Callback for when the mouse is double clicked."""



from contextlib import contextmanager
from typing import Any, Iterable, TYPE_CHECKING, Union
import uuid

from pydantic import PrivateAttr, BaseModel, Field

from himena.workflow._base import WorkflowStep
from himena.workflow._reader import (
    ProgrammaticMethod,
    LocalReaderMethod,
    RemoteReaderMethod,
    WslReaderMethod,
    UserInput,
)
from himena.workflow._command import CommandExecution, UserModification

if TYPE_CHECKING:
    from himena.types import WidgetDataModel
    from himena.mock import MainWindowMock
    from himena.mock.widget import MockWidget
    from himena.widgets import SubWindow

WorkflowStepType = Union[
    ProgrammaticMethod,
    LocalReaderMethod,
    RemoteReaderMethod,
    WslReaderMethod,
    CommandExecution,
    UserModification,
    UserInput,
]


def _make_mock_main_window():
    from himena.mock import MainWindowMock
    from himena._app_model import get_model_app
    from himena.style import default_style

    return MainWindowMock(get_model_app("."), default_style())


class Workflow(BaseModel):
    """Container of WorkflowStep instances.

    The data structure of a workflow is a directed acyclic graph. Each node is a
    WorkflowStep instance, and the edges are defined inside each CommandExecution
    instance. Each node is tagged with a unique ID named `id`, which is used as a
    mathematical identifier for the node.
    """

    steps: list[WorkflowStepType] = Field(default_factory=list)
    _mock_main_window: "MainWindowMock | None" = PrivateAttr(default=None)

    def id_to_index_map(self) -> dict[uuid.UUID, int]:
        return {step.id: i for i, step in enumerate(self.steps)}

    def filter(self, step: uuid.UUID) -> "Workflow":
        """Return another workflow that only contains the ancestors of the given ID.

        For example, if the workflow is:

        ```
        W0 -> W1 -> W2 -> W3
              ^
              |
        W2 ---+
        ```

        and the given ID is W1, then the returned workflow will be:

        ```
        W0 -> W1
              ^
              |
        W2 ---+
        ```

        """
        indices = sorted(self._get_ancestors(step)[0])
        out = Workflow(steps=[self.steps[i] for i in indices])
        # NOTE: do not update, share the reference
        out._mock_main_window = self._mock_main_window
        return out

    def __getitem__(self, index: int) -> WorkflowStep:
        return self.steps[index]

    def last(self) -> WorkflowStep | None:
        """Get the last step if exists."""
        if len(self.steps) == 0:
            return None
        return self.steps[-1]

    def last_id(self) -> uuid.UUID:
        """Get the ID of the last step."""
        if step := self.last():
            return step.id
        raise ValueError("Workflow is empty.")

    def deep_copy(self) -> "Workflow":
        return Workflow(steps=[step.model_copy() for step in self.steps])

    def index_for_id(self, id: uuid.UUID) -> int:
        for index, step in enumerate(self.steps):
            if step.id == id:
                return index
        raise ValueError(f"Workflow with id {id} not found.")

    def step_for_id(self, id: uuid.UUID) -> WorkflowStep:
        return self.steps[self.index_for_id(id)]

    def window_for_id(self, id: uuid.UUID) -> "SubWindow[MockWidget]":
        """Get the sub-window for the given ID."""
        if (main := self._mock_main_window) and (win := main.window_for_id(id)):
            return win
        step = self.step_for_id(id)
        model = step.get_model(self)
        if main := self._mock_main_window:
            win = main.add_data_model(model)
            win._identifier = id
            return win
        raise ValueError("Window input cannot be resolved in this context.")

    def model_for_id(self, id: uuid.UUID) -> "WidgetDataModel":
        """Get the widget data model for the given ID."""
        if (main := self._mock_main_window) and (win := main.window_for_id(id)):
            return win.to_model()
        step = self.step_for_id(id)
        model = step.get_model(self)
        if main := self._mock_main_window:
            win = main.add_data_model(model)
            win._identifier = id
        return model

    def __iter__(self):
        return iter(self.steps)

    def __len__(self) -> int:
        return len(self.steps)

    def with_step(
        self,
        step: WorkflowStepType,
    ) -> "Workflow":
        """Return a new workflow with the given step added."""
        if not isinstance(step, WorkflowStep):
            raise ValueError("Expected a WorkflowStep instance.")
        # The added step is always a unique node.
        return Workflow(steps=self.steps + [step])

    def compute(
        self,
        process_output: bool = True,
        metadata: Any | None = None,
    ) -> "WidgetDataModel":
        """Compute the last node in the workflow.

        Parameters
        ----------
        process_output : bool, optional
            Whether to process the output.
        metadata : Any, optional
            If given, metadata of the output will be overridden by this value.
        """
        with self._cache_context():
            out = self[-1].get_model(
                self,
                force_process_output=process_output,
                metadata=metadata,
            )
        return out

    @contextmanager
    def _cache_context(self):
        """Cache the intermediate results in this context.

        For example, if the workflow is `A -> B0`, `A -> B1`, `B0, B1 -> C`, then
        the result of `A` will be cached and reused when computing `B0` and `B1`.
        """
        was_none = self._mock_main_window is None
        if was_none:
            self._mock_main_window = _make_mock_main_window()
        try:
            yield
        finally:
            if was_none:
                self._mock_main_window.clear()
                self._mock_main_window = None

    @classmethod
    def concat(cls, workflows: Iterable["Workflow"]) -> "Workflow":
        """Concatenate multiple workflows and drop duplicate nodes based on the ID."""
        nodes: list[WorkflowStep] = []
        id_found: set[uuid.UUID] = set()
        for workflow in workflows:
            for node in workflow:
                if node.id in id_found:
                    continue
                id_found.add(node.id)
                nodes.append(node)
        return Workflow(steps=nodes)

    def replace(
        self,
        step: uuid.UUID,
        new: "WorkflowStep | Workflow",
    ) -> "Workflow":
        """Replace the step of the given ID with the new step."""
        indices, index = self._get_ancestors(step, exclude_me=True)
        new_steps = self.steps.copy()
        indices_to_remove = sorted(indices, reverse=True)
        for i in indices_to_remove:
            new_steps.pop(i)
        index_shifted = index - len([i for i in indices_to_remove if i < index])
        if isinstance(new, WorkflowStep):
            new_steps[index_shifted] = new
            new.id = step
        elif isinstance(new, Workflow):
            insert = new.steps.copy()
            if len(insert) == 0:
                raise ValueError("Input workflow is empty.")
            insert[-1] = insert[-1].model_copy(update={"id": step})
            new_steps = (
                new_steps[:index_shifted] + insert + new_steps[index_shifted + 1 :]
            )
        else:
            raise TypeError(f"Expected a WorkflowStep or Workflow, got {type(new)}.")
        return Workflow(steps=new_steps)

    def replace_with_input(
        self,
        step: uuid.UUID,
        how: str = "model",
    ) -> "Workflow":
        """Replace the step of the given ID with a runtime input."""
        new_step = UserInput(how=how)  # TODO: restrict input type
        return self.replace(step, new_step)

    def _get_ancestors(
        self,
        step: uuid.UUID,
        exclude_me: bool = False,
    ) -> tuple[set[int], int]:
        id_to_index_map = self.id_to_index_map()
        index = id_to_index_map[step]

        indices = {index}
        ancestors = [self.steps[index]]
        while ancestors:
            current = ancestors.pop()
            for id_ in current.iter_parents():
                idx = id_to_index_map[id_]
                if idx in indices:
                    continue
                indices.add(idx)
                ancestors.append(self.steps[idx])
        if exclude_me:
            indices.remove(index)
        return indices, index


def compute(
    workflows: list[Workflow],
    metadata_overrides: list | None = None,
) -> list["WidgetDataModel | Exception"]:
    """Compute all the workflow with the shared cache."""
    if len(workflows) == 0:
        return []
    _global_ui = _make_mock_main_window()
    results: list["WidgetDataModel"] = []
    all_workflows = Workflow.concat(workflows)
    # share the cache
    for workflow in workflows:
        workflow._mock_main_window = _global_ui
    # normalize metadata
    if metadata_overrides is None:
        metadata_overrides = [None] * len(workflows)
    with all_workflows._cache_context():
        for workflow, meta in zip(workflows, metadata_overrides, strict=True):
            try:
                result = workflow.compute(process_output=False, metadata=meta)
            except Exception as e:
                result = e
            results.append(result)
    _global_ui.clear()
    for workflow in workflows:
        workflow._mock_main_window = None
    return results


def is_reproducible(workflows: list[Workflow]) -> list[bool]:
    if len(workflows) == 0:
        return []
    _global_cache: dict[uuid.UUID, bool] = {}

    def _is_reproducible(
        step: WorkflowStep,
        id_to_index_map: dict[uuid.UUID, int],
    ) -> bool:
        parents = list(step.iter_parents())
        if len(parents) == 0:
            return isinstance(step, LocalReaderMethod)
        for parent in parents:
            if parent not in _global_cache:
                idx = id_to_index_map[parent]
                _global_cache[parent] = _is_reproducible(
                    workflows[idx], id_to_index_map
                )
            rep = _global_cache[parent]
            if not rep:
                return False
        return True

    results: list[bool] = []
    for workflow in workflows:
        results.append(all(_is_reproducible(step) for step in workflow.steps))
    return results
